from ..guestagent import GuestAgent, ClientError
import mock
import pytest
import socket
import StringIO


@pytest.fixture
def ga(monkeypatch):
    ga = GuestAgent('testvm', .1)
    ga.file = StringIO.StringIO()
    ga.client = mock.MagicMock()
    randint = mock.Mock(return_value=87643)
    monkeypatch.setattr('random.randint', randint)
    return ga


def test_ga_read(ga):
    ga.file = StringIO.StringIO('{"return": 17035}\n')
    assert 17035 == ga.read()


def test_ga_read_error(ga):
    ga.file = StringIO.StringIO('{"return": 0, "error": "test failure"}\n')
    with pytest.raises(ClientError):
        ga.read()


def test_ga_sync_immediate(ga):
    ga.file = StringIO.StringIO('{"return": 87643}\n')
    ga.sync()
    assert True


def test_ga_sync_retry(ga):
    ga.file = StringIO.StringIO('{"return": 2}\n{"return": 87643}\n')
    ga.sync()
    assert True


def test_ga_sync_too_often(ga):
    ga.file = StringIO.StringIO("""\
{"return": 2}
{"return": 3}
{"return": 4}
{"return": 4}
{"return": 4}
{"return": 4}
{"return": 4}
{"return": 4}
{"return": 4}
{"return": 4}
{"return": 4}
{"return": 4}
{"return": 4}
{"return": 4}
{"return": 4}
{"return": 4}
{"return": 4}
{"return": 4}
{"return": 4}
{"return": 4}
{"return": 4}
{"return": 4}
{"return": 87643}
""")
    with pytest.raises(ClientError):
        ga.sync()


def test_ga_contextmgr(ga, monkeypatch, tmpdir):
    monkeypatch.setattr(socket, 'socket', mock.MagicMock(socket.socket))
    with open(str(tmpdir / 'socket'), 'w') as f:
        f.write('{"return": 87643}\n')
    f = open(str(tmpdir / 'socket'), 'r')
    socket.socket().makefile.return_value = f
    with ga as g:
        assert g.machine == 'testvm'



import time


class TimeoutError(RuntimeError):
    pass


class TimeOut(object):

    _now = time.time

    def __init__(self, timeout, interval=1, raise_on_timeout=False,
                 log=None):
        self.cutoff = self._now() + timeout
        self.interval = interval
        self.timed_out = False
        self.first = True
        self.raise_on_timeout = raise_on_timeout
        self.log = log

    @property
    def remaining(self):
        return int(self.cutoff - self._now())

    def tick(self):
        """Perform a `tick` for this timeout.

        Returns True if we should keep going or False if not.

        Instead of returning False this can raise an exception
        if raise_on_timeout is set.

        """
        remaining = self.remaining  # make atomic within this function
        self.timed_out = remaining <= 0

        if self.timed_out:
            if self.raise_on_timeout:
                raise TimeoutError()
            else:
                return False

        if self.first:
            self.first = False
        else:
            if self.log:
                self.log.debug(
                    'waiting', interval=int(self.interval),
                    remaining=remaining)
            time.sleep(self.interval)

        return True



"""High-level handling of Ceph volumes.

We expect Ceph Python bindings to be present in the system site packages.
"""

from ..sysconfig import sysconfig
from ..util import cmd, log
from .volume import Volume
import hashlib
import rados


class Ceph(object):

    # Attributes on this class can be overriden in a controlled fashion
    # from the sysconfig module. See __init__(). The defaults are here to
    # support testing.

    CREATE_VM = None

    def __init__(self, cfg):
        # Update configuration values from system or test config.
        self.__dict__.update(sysconfig.ceph)
        self.log = log.bind(subsystem='ceph', machine=cfg['name'])

        self.cfg = cfg
        self.rados = None
        self.ioctx = None
        self.root = None
        self.swap = None
        self.tmp = None
        self.volumes = []
        self.pool = self.cfg['rbd_pool'].encode('ascii')

    def __enter__(self):
        # Not sure whether it makes sense that we configure the client ID
        # without 'client.': qemu doesn't want to see this, whereas the
        # Rados binding does ... :/
        self.log.debug('connect-rados')
        self.rados = rados.Rados(
            conffile=self.CEPH_CONF,
            name='client.' + self.CEPH_CLIENT)
        self.rados.connect()

        self.ioctx = self.rados.open_ioctx(self.pool)

        volume_prefix = self.cfg['name'].encode('ascii')
        self.root = Volume(self, volume_prefix + '.root', 'root')
        self.swap = Volume(self, volume_prefix + '.swap', 'swap')
        self.tmp = Volume(self, volume_prefix + '.tmp', 'tmp')

        self.volumes = [self.root, self.swap, self.tmp]

    def __exit__(self, exc_value, exc_type, exc_tb):
        for vol in self.volumes:
            if vol._image:
                vol._image.close()
        self.ioctx.close()
        self.rados.shutdown()

    def start(self, enc_data, generation):
        self.ensure_root_volume()
        self.ensure_tmp_volume(enc_data, generation)
        self.ensure_swap_volume()

    def stop(self):
        self.unlock()

    def ensure_root_volume(self):
        self.log.info('ensure-root')
        if not self.root.exists():
            self.log.info('create-vm')
            cmd(self.CREATE_VM.format(**self.cfg), self.log)
        self.root.lock()

    def ensure_swap_volume(self):
        self.log.info('ensure-swap')
        self.swap.ensure_presence(self.cfg['swap_size'])
        self.swap.lock()
        self.swap.ensure_size(self.cfg['swap_size'])
        with self.swap.mapped():
            self.swap.mkswap()

    def ensure_tmp_volume(self, enc_data, generation):
        self.log.info('ensure-tmp')
        self.tmp.ensure_presence(self.cfg['tmp_size'])
        self.tmp.lock()
        self.tmp.ensure_size(self.cfg['tmp_size'])
        with self.tmp.mapped():
            self.tmp.mkfs()
            self.tmp.seed(enc_data, generation)

    def locks(self):
        for vol in self.volumes:
            status = vol.lock_status()
            if not status:
                continue
            yield vol.name, status[1]

    def is_unlocked(self):
        """Returns True if no volume is locked."""
        return all(not vol.lock_status() for vol in self.volumes)

    def locked_by_me(self):
        """Returns True if CEPH_LOCK_HOST holds locks for all volumes."""
        try:
            return all(v.lock_status()[1] == self.CEPH_LOCK_HOST
                       for v in self.volumes)
        except TypeError:  # status[1] not accessible
            return False

    def locked_by(self):
        """Returns a hostname holding all locks or None if not locked.

        Raises ValueError if not all locks are held by same owner.

        """
        lock_owners = set(
            v.lock_status()[1] for v in self.volumes if v.lock_status())
        if not lock_owners:
            return None
        if len(lock_owners) != 1:
            raise ValueError("Multiple lock owners: {}".format(lock_owners))
        return lock_owners.pop()

    def lock(self):
        for vol in self.volumes:
            vol.lock()

    def unlock(self):
        """Remove all of *our* volume locks.

        We try to agressively get rid of as many locks as we can, but propagate
        an exception if it occurs.

        This leaves other hosts' locks in place.
        """
        exception = False
        for vol in self.volumes:
            try:
                vol.unlock()
            except Exception:
                vol.log.warning('unlock-failed', exc_info=True)
                exception = True
        if exception:
            raise RuntimeError(
                "Failed to unlock all locks. See log for specific exceptions.")

    def force_unlock(self):
        for vol in self.volumes:
            vol.unlock(force=True)

    def auth_cookie(self):
        """This is a cookie that can be used to validate that a party
        has access to Ceph.

        Used to authenticate migration requests.
        """
        c = hashlib.sha1()
        for vol in self.volumes:
            status = [vol.name]
            lock = vol.lock_status()
            if lock:
                status.extend(lock)
            c.update('\0'.join(status) + '\0')
        return c.hexdigest()



import ConfigParser
import os.path


class SysConfig(object):
    """A global config state registry.

    This is used to manage system-specific configuration overrides
    for values used in various places within fc.qemu.

    This provides the code for loading all those options from a central
    config file and to allow tests overriding those values gracefully.
    """

    def __init__(self):
        self.qemu = {}
        self.ceph = {}
        self.agent = {}
        self.cp = None

    def read_config_files(self):
        """Tries to open fc-qemu.conf at various location."""
        self.cp = ConfigParser.SafeConfigParser()
        self.cp.read(os.path.dirname(__file__) + '/default.conf')
        self.cp.read('/etc/qemu/fc-qemu.conf')
        if 'qemu' not in self.cp.sections():
            raise RuntimeError('error while reading config file: '
                               'section [qemu] not found')

    def load_system_config(self):
        self.read_config_files()

        self.qemu['migration_address'] = self.cp.get(
            'qemu', 'migration-address')
        self.qemu['require_kvm'] = bool(self.cp.get('qemu', 'accelerator'))
        self.qemu['vnc'] = self.cp.get('qemu', 'vnc')
        self.qemu['max_downtime'] = self.cp.getfloat('qemu', 'max-downtime')
        self.qemu['vm_max_total_memory'] = self.cp.getint(
            'qemu', 'vm-max-total-memory')
        self.qemu['vm_expected_overhead'] = self.cp.getint(
            'qemu', 'vm-expected-overhead')

        self.qemu['throttle_by_pool'] = tbp = {}
        for pool, iops in self.cp.items('qemu-throttle-by-pool'):
            tbp[pool] = int(iops)

        # Consul
        self.agent['consul_token'] = self.cp.get('consul', 'access-token')
        self.agent['consul_event_threads'] = self.cp.getint(
            'consul', 'event-threads')

        # Qemu
        self.agent['accelerator'] = self.cp.get('qemu', 'accelerator')
        self.agent['machine_type'] = self.cp.get('qemu', 'machine-type')
        self.agent['migration_ctl_address'] = self.cp.get(
            'qemu', 'migration-ctl-address')
        self.agent['binary_generation'] = self.cp.getint(
            'qemu', 'binary-generation')
        self.agent['timeout_graceful'] = self.cp.getint(
            'qemu', 'timeout-graceful')
        self.agent['vhost'] = self.cp.getboolean('qemu', 'vhost')

        # Ceph
        self.agent['this_host'] = self.cp.get('ceph', 'lock_host')
        self.agent['ceph_id'] = self.cp.get('ceph', 'client-id')

        self.ceph['CEPH_CLIENT'] = self.cp.get('ceph', 'client-id', 'admin')
        self.ceph['CEPH_CLUSTER'] = self.cp.get('ceph', 'cluster', 'ceph')
        self.ceph['CEPH_CONF'] = self.cp.get('ceph', 'ceph-conf')
        self.ceph['CEPH_LOCK_HOST'] = self.cp.get('ceph', 'lock_host')
        self.ceph['CREATE_VM'] = self.cp.get('ceph', 'create-vm')
        self.ceph['MKFS_XFS'] = self.cp.get('ceph', 'mkfs-xfs')
        self.ceph['MKFS_EXT4'] = self.cp.get('ceph', 'mkfs-ext4')


sysconfig = SysConfig()



from ..outgoing import Outgoing
import pytest
import mock


@pytest.fixture
def outgoing():
    o = Outgoing(mock.MagicMock())
    o.target = mock.MagicMock()
    return o


def test_prefer_remote_rescue(outgoing):
    outgoing.rescue()
    assert outgoing.target.rescue.called is True
    assert outgoing.agent.qemu.destroy.called is True


def test_request_remote_destroy_if_remote_rescue_fails(outgoing):
    outgoing.target.rescue.side_effect = RuntimeError('boom')
    outgoing.rescue()
    assert outgoing.target.destroy.called is True
    assert outgoing.agent.qemu.destroy.called is False
    assert outgoing.agent.ceph.lock.called is True



"""Low-level handling of Ceph volumes.

This module contains the Volume, Snapshot, and Image classes. Volume
represents an RBD volume that is not a snapshot. Image is an abstract
base class for both volumes and snapshots.
"""

from ..util import remove_empty_dirs, cmd
import contextlib
import json
import os
import os.path as p
import rbd
import time


class Image(object):
    """Abstract base class for all images (volumes and snapshots)."""

    device = None
    mountpoint = None

    def __init__(self, ceph, name):
        self.ceph = ceph
        self.ioctx = ceph.ioctx
        self.name = name
        self.rbd = rbd.RBD()

    def __str__(self):
        return self.fullname

    @property
    def rbdimage(self):  # pragma: no cover
        raise NotImplementedError

    @property
    def fullname(self):  # pragma: no cover
        raise NotImplementedError

    @property
    def part1dev(self):
        if not self.device:
            return None
        return self.device + '-part1'

    def map(self):
        if self.device is not None:
            return
        self.cmd('rbd -c "{}" --id "{}" map "{}"'.format(
                 self.ceph.CEPH_CONF, self.ceph.CEPH_CLIENT, self.fullname))
        time.sleep(0.1)
        self.device = '/dev/rbd/' + self.fullname

    def unmap(self):
        if self.device is None:
            return
        self.cmd('rbd -c "{}" --id "{}" unmap "{}"'.format(
                 self.ceph.CEPH_CONF, self.ceph.CEPH_CLIENT, self.device))
        self.device = None

    @contextlib.contextmanager
    def mapped(self):
        """Maps the image to a block device and yields the device name."""
        self.map()
        try:
            yield self.device
        finally:
            self.unmap()

    def mount(self):
        if self.mountpoint is not None:
            return
        if not self.device:
            raise RuntimeError('image must be mapped before mounting',
                               self.fullname)
        mountpoint = '/mnt/rbd/{}'.format(self.fullname)
        try:
            os.makedirs(mountpoint)
        except OSError:  # pragma: no cover
            pass
        self.cmd('mount "{}" "{}"'.format(self.part1dev, mountpoint))
        self.mountpoint = mountpoint

    def umount(self):
        if self.mountpoint is None:
            return
        cmd('umount "{}"'.format(self.mountpoint), log=self.log)
        remove_empty_dirs(self.mountpoint)
        self.mountpoint = None

    @contextlib.contextmanager
    def mounted(self):
        """Mounts the image and yields mountpoint."""
        must_unmap = False
        if not self.device:
            self.map()
            must_unmap = True
        self.mount()
        try:
            yield self.mountpoint
        finally:
            self.umount()
            if must_unmap:
                self.unmap()


class Snapshots(object):
    """Container for all snapshots of a Volume."""

    def __init__(self, volume):
        self.vol = volume
        self.log = self.vol.log

    def __iter__(self):
        """Iterate over all existing snapshots."""
        return (Snapshot(self.vol, s['name'], s['id'], s['size'])
                for s in self._list_snaps())

    def __len__(self):
        return len(list(self._list_snaps()))

    def __getitem__(self, key):
        for s in self._list_snaps():
            if key == s['name']:
                return Snapshot(self.vol, s['name'], s['id'], s['size'])
        raise KeyError(key)

    def _list_snaps(self):
        try:
            return self.vol.rbdimage.list_snaps()
        except rbd.ImageNotFound:
            return []

    def purge(self):
        """Remove all snapshots."""
        for snapshot in self:
            self.log.info('remove-snapshot',
                          volum=self.vol.fullname,
                          snapshot=snapshot.name)
            snapshot.remove()

    def create(self, snapname):
        self.log.info('create-snapshot',
                      volume=self.vol.fullname,
                      snapshot=snapname)
        self.vol.rbdimage.create_snap(snapname)


class Snapshot(Image):
    """Single snapshot of a Volume."""

    def __init__(self, volume, snapname, id, snapsize):
        super(Snapshot, self).__init__(volume.ceph, volume.name)
        self.vol = volume
        self.snapname = snapname
        self.id = id
        self.size = snapsize
        self.log = volume.log.bind(snapshot=snapname)
        self.cmd = lambda cmdline: cmd(cmdline, self.log)

    @property
    def rbdimage(self):
        return rbd.Image(self.ioctx, self.name, self.snapname)

    @property
    def fullname(self):
        return self.ioctx.name + '/' + self.name + '@' + self.snapname

    def remove(self):
        """Destroy myself."""
        self.log.info('remove-snapshot')
        self.vol.rbdimage.remove_snap(self.snapname)


class Volume(Image):
    """RBD Volume interface.

    A volume has a name (shown in `rbd ls`) and a filesystem label. The
    latter is only used for mkfs/mkswap. A volume may have a set of
    snapshots accessible via the `snapshots` container object.
    """

    MKFS_CMD = {
        'xfs': ('mkfs.xfs {options} -L "{label}" "{device}"'),
        'ext4': ('mkfs.ext4 {options} -L "{label}" "{device}" '
                 '&& tune2fs -e remount-ro "{device}"')
    }

    def __init__(self, ceph, name, label):
        super(Volume, self).__init__(ceph, name)
        self.label = label
        self.log = ceph.log.bind(volume=self.fullname)
        self.cmd = lambda cmdline: cmd(cmdline, log=self.log)
        self.snapshots = Snapshots(self)
        self.locked_by_me = False
        self._image = None

    @property
    def rbdimage(self):
        if not self._image:
            self._image = rbd.Image(self.ioctx, self.name)
        return self._image

    @property
    def fullname(self):
        return self.ioctx.name + '/' + self.name

    @property
    def size(self):
        """Image size in Bytes."""
        return self.rbdimage.size()

    def exists(self):
        return self.name in self.rbd.list(self.ioctx)

    def ensure_presence(self, size=1024):
        if self.name in self.rbd.list(self.ioctx):
            return
        self.rbd.create(self.ioctx, self.name, size)

    def ensure_size(self, size):
        if self.size == size:
            return
        self.rbdimage.resize(size)

    def lock(self):
        self.log.info('lock')
        retry = 3
        while retry:
            try:
                self.rbdimage.lock_exclusive(self.ceph.CEPH_LOCK_HOST)
                self.locked_by_me = True
            except rbd.ImageExists:
                # This client and cookie already locked this. This is
                # definitely fine.
                return
            except rbd.ImageBusy:
                # Maybe the same client but different cookie. We're fine with
                # different cookies - ignore this. Must be same client, though.
                status = self.lock_status()
                if status is None:
                    # Someone had the lock but released it in between.
                    # Lets try again
                    retry -= 1
                    continue
                if status[1] == self.ceph.CEPH_LOCK_HOST:
                    # That's locked for us already. We just re-use
                    # the existing lock.
                    return
                # Its locked for some other client.
                self.log.error('assume-lock-failed', competing=status)
                raise
        raise rbd.ImageBusy(
            'Could not acquire lock - tried multiple times. '
            'Someone seems to be racing me.')

    def lock_status(self):
        """Return None if not locked and (client_id, lock_id) if it is."""
        try:
            lockers = self.rbdimage.list_lockers()
        except rbd.ImageNotFound:
            return None
        if not lockers:
            return
        # For some reasons list_lockers may just return an empty list instead
        # of a dict. Say what?
        lockers = lockers['lockers']
        if not len(lockers) == 1:
            raise NotImplementedError("I'm not prepared for shared locks")
        client_id, lock_id, addr = lockers[0]
        return client_id, lock_id

    def unlock(self, force=False):
        locked_by = self.lock_status()
        if not locked_by:
            return
        client_id, lock_id = locked_by
        if self.locked_by_me:
            self.log.info('unlock')
            self.rbdimage.unlock(lock_id)
            self.locked_by_me = False
            return

        def _break_lock():
            try:
                self.rbdimage.break_lock(client_id, lock_id)
            except rbd.InvalidArgument:
                if client_id == 'client.?':
                    self.log.warn('buggy-lock', action='low-level-delete')
                    # This looks a lot like a broken hammer installation.
                    # Be cruel.
                    img_id = self.ioctx.read('rbd_id.{}'.format(self.name)).strip()
                    img_id = img_id[4:]
                    self.ioctx.rm_xattr('rbd_header.{}'.format(img_id), 'lock.rbd_lock')

        if lock_id == self.ceph.CEPH_LOCK_HOST:
            # This host owns this lock but from a different connection. This
            # means we have to break it.
            self.log.info('unlock')
            _break_lock()
        # We do not own this lock: need to explicitly ask for breaking.
        elif force:
            self.log.info('break-lock')
            _break_lock()

    def mkswap(self):
        """Creates a swap partition. Requires the volume to be mappped."""
        assert self.device, 'volume must be mapped first'
        self.cmd('mkswap -f -L "{}" "{}"'.format(self.label, self.device))

    def mkfs(self, fstype='xfs', gptbios=False):
        self.log.debug('create-fs', type=fstype)
        assert self.device, 'volume must be mapped first'
        self.cmd('sgdisk -o "{}"'.format(self.device))
        self.cmd('sgdisk -a 8192 -n 1:8192:0 -c "1:{}" -t 1:8300 '
                 '"{}"'.format(self.label, self.device))
        if gptbios:
            self.cmd('sgdisk -n 2:2048:+1M -c 2:gptbios -t 2:EF02 "{}"'.format(
                     self.device))
        self.cmd('partprobe {}'.format(self.device))
        time.sleep(0.2)
        while not p.exists(self.part1dev):  # pragma: no cover
            time.sleep(0.1)
        options = getattr(self.ceph, 'MKFS_' + fstype.upper())
        self.cmd(self.MKFS_CMD[fstype].format(
                 options=options, device=self.part1dev, label=self.label))

    def seed(self, enc, generation):
        self.log.info('seed')
        with self.mounted() as target:
            os.chmod(target, 0o1777)
            fc_data = p.join(target, 'fc-data')
            os.mkdir(fc_data)
            os.chmod(fc_data, 0o750)
            with open(p.join(fc_data, 'enc.json'), 'w') as f:
                os.fchmod(f.fileno(), 0o640)
                json.dump(enc, f)
                f.write('\n')
            generation_marker = p.join(
                fc_data, 'qemu-binary-generation-booted')
            with open(generation_marker, 'w') as f:
                f.write(str(generation) + '\n')



from .timeout import TimeOut
from .exc import ConfigChanged
import pprint
import random
import xmlrpclib


class Outgoing(object):

    migration_exitcode = None
    target = None
    cookie = None

    # How long to wait until we discover an inmigrate service?
    # This should happen relatively fast, but just in case we'll keep a high
    # timeout. The agents should be spawned relatively quickly everywhere.
    connect_timeout = 60 * 60  # 1 hour

    # How long to wait until we get a migration lock?
    # This can happen quite slowly as during busy times migration may be slow
    # and some large and busy VMs can take 1-2hours (on Gigabit). If this adds
    # up over multiple hosts I'm giving a grace period of up to 12 hours here.
    migration_lock_timeout = 12 * 60 * 60  # 12 hours.

    def __init__(self, agent):
        self.agent = agent
        self.log = agent.log
        self.name = agent.name
        self.consul = agent.consul

    def __call__(self):
        self.cookie = self.agent.ceph.auth_cookie()
        with self:
            self.connect()
            self.acquire_migration_locks()
            self.transfer_ceph_locks()
            self.migrate()

        return self.migration_exitcode

    # The context manager is suitable to express "I am going to
    # interact with this VM: please prepare connections to
    # the according backends and make sure to clean up after me.""
    def __enter__(self):
        return self

    def __exit__(self, _exc_type, _exc_value, _traceback):
        if _exc_type is None:
            # We're on our happy path. Release the migration lock
            # optimistically and only in this case. Otherwise the
            # error reporting for why the migration has failed will
            # be screwed. And the negative path will exit the process
            # and eventually give up locks anyway.
            try:
                self.agent.qemu.release_migration_lock()
            except Exception:
                pass
            try:
                self.target.release_migration_lock(self.cookie)
            except Exception:
                pass

            self.migration_exitcode = 0

        else:
            self.migration_exitcode = 1
            try:
                # XXX silence timeout errors here to avoid unnecessary
                # traceback output
                self.log.exception(
                    'migration-failed', action='rescue', exc_info=True)
                self.rescue()
                return True  # swallow exception
            except:  # noqa
                # Purposeful bare except: try really hard to kill
                # our VM.
                self.log.exception(
                    'rescue-failed', exc_info=True, action='destroy')
                self.destroy()

    def locate_inmigrate_service(self):
        service_name = 'vm-inmigrate-' + self.name
        timeout = TimeOut(
            self.connect_timeout, interval=3, raise_on_timeout=True,
            log=self.log)
        self.log.info('locate-inmigration-service')
        while timeout.tick():
            if self.agent.has_new_config():
                raise ConfigChanged()
            candidates = self.consul.catalog.service(service_name)
            if len(candidates) > 1:
                self.log.warning('multiple-services-found',
                                 action='trying newest first',
                                 service=service_name)
            candidates = sorted(
                candidates, key=lambda i: i['ModifyIndex'], reverse=True)
            for candidate in candidates:
                url = 'http://{}:{}'.format(
                    candidate['ServiceAddress'], candidate['ServicePort'])
                self.log.info('located-inmigration-service', url=url)
                try:
                    target = xmlrpclib.ServerProxy(url, allow_none=True)
                    target.ping(self.cookie)
                except Exception as e:
                    self.log.info('connect', result='failed', reason=str(e))
                    # Not a viable service. Delete it.
                    # XXX This seems broken in consulate 0.6, waiting
                    # for the 1.0 release ...
                    # try:
                    #     self.consul.catalog.deregister(
                    #        node=candidate['Node'],
                    #        datacenter=candidate['Datacenter'],
                    #        service_id=candidate['ServiceID'])
                    #     self.log.debug('delete-stale-incoming-service', result='success')
                    #     pass
                    # except Exception:
                    #     self.log.exception('delete-stale-incoming-service', result='failed', exc_info=True)
                    pass
                else:
                    break
            else:
                # We did not find a viable target - continue waiting
                continue
            return target
        raise RuntimeError('failed to locate inmigrate service', service_name)

    def connect(self):
        connection = self.locate_inmigrate_service()
        self.target = connection

    def acquire_migration_locks(self):
        tries = 0
        self.log.info('acquire-migration-locks')
        timeout = TimeOut(
            self.migration_lock_timeout, interval=3, raise_on_timeout=True,
            log=self.log)
        while timeout.tick():
            # In case that there are multiple processes waiting, randomize to
            # avoid steplock retries. We us CSMA/CD-based exponential backoff,
            # timeslot 10ms but with a max of 13 instead of 16.
            # This means we'll wait up to 80s, 40s on average if everything
            # becomes really busy and we may experience lock contention.
            tries = min([tries+1, 13])
            timeout.interval = random.randint(1, 2**tries) * 0.01
            if self.agent.has_new_config():
                self.target.cancel(self.cookie)
                raise ConfigChanged()
            # Keep the remote peer alive by informing it that we're still
            # alive and working on it. Add additional grace period to our own
            # interval.
            self.target.ping(self.cookie, timeout.interval + 60)

            # Try to acquire local lock
            if self.agent.qemu.acquire_migration_lock():
                self.log.debug('acquire-local-migration-lock', result='success')
            else:
                self.log.debug('acquire-local-migration-lock', result='failure')
                continue

            # Try to acquire remote lock
            try:
                self.log.debug('acquire-remote-migration-lock')
                # We got our lock, now ask the remote side:
                if not self.target.acquire_migration_lock(self.cookie):
                    self.log.debug(
                        'acquire-remote-migration-lock', result='failure')
                    self.agent.qemu.release_migration_lock()
                    continue
                self.log.debug(
                    'acquire-remote-migration-lock', result='success')
            except Exception:
                self.log.exception(
                    'acquire-remote-migration-lock', result='failure', exc_info=True)
                self.agent.qemu.release_migration_lock()
                continue
            # Reset the hard timeout to regular ping timeouts.
            self.target.ping(self.cookie)
            break

    def transfer_ceph_locks(self):
        self.agent.ceph.unlock()
        self.target.acquire_ceph_locks(self.cookie)

    def migrate(self):
        """Actually move VM between hosts."""
        args, config = self.agent.qemu.get_running_config()
        self.log.info('prepare-remote-environment')
        migration_address = self.target.prepare_incoming(
            self.cookie, args, config)
        self.log.info('start-migration', target=migration_address)
        self.agent.qemu.migrate(migration_address)
        try:
            for stat in self.agent.qemu.poll_migration_status():
                remaining = stat['ram']['remaining'] if 'ram' in stat else 0
                mbps = stat['ram']['mbps'] if 'ram' in stat else '-'
                self.log.info('migration-status',
                              status=stat['status'],
                              remaining='{0:,d}'.format(remaining),
                              mbps=mbps,
                              output=pprint.pformat(stat))
                self.target.ping(self.cookie)
        except Exception:
            self.log.exception('error-waiting-for-migration', exc_info=True)
            raise

        status = self.agent.qemu.qmp.command('query-status')
        assert not status['running'], status
        assert status['status'] == 'postmigrate', status
        self.log.info('finish-migration')
        try:
            self.target.finish_incoming(self.cookie)
        except Exception:
            self.log.exception('error-finish-remote', exc_info=True)
        self.agent.qemu.destroy()

    def rescue(self):
        """Outgoing rescue: try to rescue the remote side first."""
        if self.target is not None:
            try:
                self.log.info('rescue-remote')
                self.target.rescue(self.cookie)
                self.target.finish_incoming(self.cookie)
                self.log.info('rescue-remote-success', action='destroy local')
                self.destroy()
                # We managed to rescue on the remote side - hooray!
                self.migration_exitcode = 0
                return
            except Exception:
                self.log.exception('rescue-remote-failed', exc_info=True,
                                   action='destroy remote')
                try:
                    self.target.destroy(self.cookie)
                except Exception:
                    self.log.exception('destroy-remote-failed', exc_info=True)
        try:
            self.log.info('continue-locally')
            self.agent.ceph.lock()
            assert self.agent.qemu.is_running()
        except Exception:
            self.log.exception('continue-locally', exc_info=True,
                               result='failed', action='destroy local')
            self.destroy()
        else:
            self.log.info('continue-locally',
                          result='success')

    def destroy(self):
        self.agent.qemu.destroy()
        self.agent.qemu.clean_run_files()



import re
import difflib


class Report(object):

    matches = None

    def __init__(self):
        self.lines = []

    def matched(self, line):
        self.lines.append((True, line))

    def nonmatched(self, line, match):
        self.lines.append((False, line, match))

    @property
    def diff(self):
        result = ['']
        d = difflib.Differ()
        for line in self.lines:
            if line[0]:
                result.append('  ' + line[1])
            else:
                if line[2] is None:
                    result.append('+ ' + line[1])
                elif line[1] is None:
                    result.append('- ' + line[2])
                else:
                    diffed = d.compare([line[2]], [line[1]])
                    diffed = [x.rstrip('\n') for x in diffed]
                    result.extend(diffed)
        result = list(filter(str.strip, result))
        return result

    @property
    def is_ok(self):
        return all(x[0] for x in self.lines)


def match(pattern, line):
    pattern = pattern.replace('\t', ' ' * 8)
    line = line.replace('\t', ' ' * 8)
    pattern = re.escape(pattern)
    pattern = pattern.replace('\.\.\.', '.+?')
    pattern = re.compile('^' + pattern + '$')
    return pattern.match(line)


class Ellipsis(object):

    # other = other.replace('\t', ' '*8) oder allgemein white-space unsensibel
    # multi-line support

    def __init__(self, ellipsis):
        self.patterns = ellipsis.split('\n')

    def compare(self, lines):
        report = Report()

        patterns = self.patterns[:]
        # Keep track of whether we're on a multi-line ellipsis.
        multiline = False
        pattern = None

        for line in lines.split('\n'):
            # Select next applicable pattern.
            if multiline:
                report.matched(line)
                if pattern and match(pattern, line):
                    multiline = False
            else:
                if patterns:
                    pattern = patterns.pop(0)
                    if pattern == '...':
                        multiline = True
                        if patterns:
                            pattern = patterns.pop(0)
                        else:
                            pattern = None
                        report.matched(line)
                        continue
                    if match(pattern, line):
                        report.matched(line)
                    else:
                        report.nonmatched(line, pattern)
                else:
                    report.nonmatched(line, None)

        if multiline and pattern:
            # Get the unmatched multi-line boundary pattern back
            # into the list of unconsumed patterns.
            patterns.insert(0, pattern)

        for pattern in patterns:
            report.nonmatched(None, pattern)

        return report

    def __eq__(self, other):
        assert isinstance(other, str)
        report = self.compare(other)
        return report.is_ok



from ..incoming import IncomingServer, IncomingAPI
from ..exc import MigrationError
import mock
import pytest


@pytest.fixture
def mock_agent():
    agent = mock.Mock()
    agent.migration_ctl_address = 'localhost:12345'
    agent.qemu = mock.Mock()
    agent.ceph = mock.Mock()
    agent.ceph.auth_cookie.return_value = '5f620fda'
    return agent


def test_prepare_should_stop_ceph_on_exception(mock_agent):
    mock_agent.qemu.inmigrate.side_effect = Exception('boom!')
    s = IncomingServer(mock_agent)
    with pytest.raises(Exception):
        s.prepare_incoming('args', 'config')
    assert mock_agent.ceph.stop.called is True


def test_rescue(mock_agent):
    s = IncomingServer(mock_agent)
    mock_agent.ceph.lock.side_effect = Exception('boom!')
    with pytest.raises(Exception):
        s.rescue()
    assert mock_agent.qemu.destroy.called is True
    assert mock_agent.ceph.unlock.called is True


@mock.patch('fc.qemu.incoming.IncomingServer')
def test_authentication_match(server):
    api = IncomingAPI(server)
    api.cookie = 'cookie1'
    # should not raise an exception
    assert api.ping('cookie1') is None


@mock.patch('fc.qemu.incoming.IncomingServer')
def test_authentication_mismatch(server):
    api = IncomingAPI(server)
    api.cookie = 'cookie1'
    with pytest.raises(MigrationError):
        assert api.ping('cookie-does-not-match') is None


def test_screen_config(mock_agent):
    s = IncomingServer(mock_agent)
    assert s.screen_config("""\
[machine]
  type = "pc-i440fx-2.5"
  iommu = "off"
  accel = "kvm"

""") == """\
[machine]
  type = "pc-i440fx-2.5"

  accel = "kvm"

"""



# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the MIT License.  See the LICENSE file in the root of this
# repository for complete details.
# The ConsoleRenderer is based on structlog.dev.ConsoleRenderer

from __future__ import absolute_import, division, print_function
from six import StringIO
import structlog
import sys
import os

try:
    import colorama
except ImportError:
    colorama = None

_MISSING = (
    "{who} requires the {package} package installed.  "
    "If you want to use the helpers from structlog.dev, it is strongly "
    "recommended to install structlog using `pip install structlog[dev]`."
)
_EVENT_WIDTH = 30  # pad the event name to so many characters


def _pad(s, l):
    """
    Pads *s* to length *l*.
    """
    missing = l - len(s)
    return s + " " * (missing if missing > 0 else 0)


if sys.stdout.isatty() and colorama:
    RESET_ALL = colorama.Style.RESET_ALL
    BRIGHT = colorama.Style.BRIGHT
    DIM = colorama.Style.DIM
    RED = colorama.Fore.RED
    BACKRED = colorama.Back.RED
    BLUE = colorama.Fore.BLUE
    CYAN = colorama.Fore.CYAN
    MAGENTA = colorama.Fore.MAGENTA
    YELLOW = colorama.Fore.YELLOW
    GREEN = colorama.Fore.GREEN
else:
    RESET_ALL = ''
    BRIGHT = ''
    DIM = ''
    RED = ''
    BACKRED = ''
    BLUE = ''
    CYAN = ''
    MAGENTA = ''
    YELLOW = ''
    GREEN = ''


class MultiOptimisticLoggerFactory(object):

    def __init__(self, **factories):
        self.factories = factories

    def __call__(self, *args):
        loggers = {k: f() for k, f in self.factories.items()}
        return MultiOptimisticLogger(loggers)


class MultiOptimisticLogger(object):

    def __init__(self, loggers):
        self.loggers = loggers

    def __repr__(self):
        return '<MultiOptimisticLogger {}>'.format(
            [repr(l) for l in self.loggers])

    def msg(self, **event_dict):
        for name, logger in self.loggers.items():
            try:
                line = event_dict.get(name)
                if line:
                    logger.msg(line)
            except Exception:
                # We're being really optimistic: we want the calling program
                # to continue even if we face huge troubles logging stuff.
                pass

    def __getattr__(self, name):
        return self.msg


def prefix(prefix, line):
    return '{}>\t'.format(prefix) + line.replace(
        '\n', '\n{}>\t'.format(prefix))


class MultiConsoleRenderer(object):
    """
    Render `event_dict` nicely aligned, in colors, and ordered with
    specific knowledge about fc.qemu structures.
    """

    LEVELS = ['exception', 'critical', 'error', 'warn', 'warning',
              'info', 'debug']

    def __init__(self, min_level, pad_event=_EVENT_WIDTH):
        self.min_level = self.LEVELS.index(min_level.lower())
        if colorama is None:
            raise SystemError(
                _MISSING.format(
                    who=self.__class__.__name__,
                    package="colorama"
                )
            )
        if sys.stdout.isatty():
            colorama.init()

        self._pad_event = pad_event
        self._level_to_color = {
            "critical": RED,
            "exception": RED,
            "error": RED,
            "warn": YELLOW,
            "warning": YELLOW,
            "info": GREEN,
            "debug": GREEN,
            "notset": BACKRED,
        }
        for key in self._level_to_color.keys():
            self._level_to_color[key] += BRIGHT
        self._longest_level = len(max(
            self._level_to_color.keys(),
            key=lambda e: len(e)
        ))

    def __call__(self, logger, method_name, event_dict):
        console_io = StringIO()
        log_io = StringIO()

        def write(line):
            console_io.write(line)
            if RESET_ALL:
                for SYMB in [RESET_ALL, BRIGHT, DIM, RED, BACKRED,
                             BLUE, CYAN, MAGENTA, YELLOW, GREEN]:
                    line = line.replace(SYMB, '')
            log_io.write(line)

        ts = event_dict.pop("timestamp", None)
        if ts is not None:
            write(
                # can be a number if timestamp is UNIXy
                DIM + str(ts) + RESET_ALL + " ")

        pid = event_dict.pop("pid", None)
        if pid is not None:
            write(
                DIM + str(pid) + RESET_ALL + " ")

        level = event_dict.pop("level", None)
        if level is not None:
            write(self._level_to_color[level] + level[0].upper() +
                  RESET_ALL + ' ')

        machine = event_dict.pop('machine', '')
        if machine:
            write(machine.ljust(20) + ' ')

        event = event_dict.pop("event")
        write(BRIGHT +
              _pad(event, self._pad_event) +
              RESET_ALL + " ")

        logger_name = event_dict.pop("logger", None)
        if logger_name is not None:
            write("[" + BLUE + BRIGHT +
                  logger_name + RESET_ALL +
                  "] ")

        output = event_dict.pop("output", None)
        args = event_dict.pop("args", None)
        stack = event_dict.pop("stack", None)
        exc = event_dict.pop("exception", None)
        write(" ".join(CYAN + key + RESET_ALL +
                       "=" +
                       MAGENTA + repr(event_dict[key]) +
                       RESET_ALL
                       for key in sorted(event_dict.keys())))

        if args is not None:
            write('\n' + DIM +
                  prefix(machine, event + ' ' + ''.join(args)) +
                  RESET_ALL)
        if output is not None:
            write('\n' + DIM + prefix(machine, output) + RESET_ALL)

        if stack is not None:
            write("\n" + prefix(machine, stack))
            if exc is not None:
                write("\n\n" + prefix(machine, "=" * 79 + "\n"))
        if exc is not None:
            write("\n" + prefix(machine, exc))

        # Filter according to the -v switch when outputting to the
        # console.
        if self.LEVELS.index(method_name.lower()) > self.min_level:
            console_io.seek(0)
            console_io.truncate()

        return {'console': console_io.getvalue(), 'file': log_io.getvalue()}


def method_to_level(logger, method_name, event_dict):
    event_dict['level'] = method_name
    return event_dict


def add_pid(logger, method_name, event_dict):
    event_dict['pid'] = os.getpid()
    return event_dict


def init_logging(verbose=True):
    log_file = open('/var/log/fc-qemu.log', 'a')
    structlog.configure(
        processors=[
            method_to_level,
            add_pid,
            structlog.processors.format_exc_info,
            structlog.processors.TimeStamper(fmt='iso', utc=False),
            MultiConsoleRenderer(min_level='debug' if verbose else 'info')
        ],
        logger_factory=MultiOptimisticLoggerFactory(
            console=structlog.PrintLoggerFactory(),
            file=structlog.PrintLoggerFactory(log_file))
    )



from ..conftest import get_log
from ..ellipsis import Ellipsis
from fc.qemu import util
from fc.qemu.agent import Agent
from fc.qemu.hazmat.qemu import Qemu
from StringIO import StringIO
import json
import os.path
import pytest

# globally overriding timeouts since _handle_consul_event creates new
# Agent/Qemu/... instances itself.
Qemu.guestagent_timeout = .01
Qemu.qmp_timeout = .01
Qemu.thaw_retry_timeout = .1


def test_no_events():
    stdin = StringIO("[]")
    Agent.handle_consul_event(stdin)
    assert util.log_data == []


def test_empty_event():
    stdin = StringIO(
        '[{"ModifyIndex": 123, "Value": "", "Key": "node/test22"}]')
    Agent.handle_consul_event(stdin)
    assert util.log_data == [
        'count=1 event=start-consul-events',
        'event=handle-key key=node/test22',
        'event=ignore-key key=node/test22 reason=empty value',
        'event=finish-consul-events']


def test_no_key_event():
    stdin = StringIO(
        '[{"ModifyIndex": 123, "Value": ""}]')
    Agent.handle_consul_event(stdin)
    assert util.log_data == [
        'count=1 event=start-consul-events',
        'event=handle-key-failed exc_info=True key=None',
        'event=finish-handle-key key=None',
        'event=finish-consul-events']


@pytest.yield_fixture
def clean_config_test22():
    targets = ['/etc/qemu/vm/test22.cfg', '/etc/qemu/vm/.test22.cfg.staging']
    for target in targets:
        if os.path.exists(target):
            os.unlink(target)
    yield
    for target in targets:
        if os.path.exists(target):
            os.unlink(target)


def test_qemu_config_change(clean_config_test22):
    util.test_log_options['show_methods'] = ['info']
    cfg = {u'classes': [u'role::appserver',
                        u'role::backupclient',
                        u'role::generic',
                        u'role::postgresql90',
                        u'role::pspdf',
                        u'role::webproxy'],
           u'consul-generation': 128,
           u'name': u'test22',
           u'parameters': {u'ceph_id': u'admin',
                           u'cores': 1,
                           u'directory_password': u'1jmGYd3dddyjpFlLqg63RHie',
                           u'directory_ring': 1,
                           u'disk': 15,
                           u'environment': u'staging',
                           u'id': 4097,
                           u'interfaces': {},
                           u'kvm_host': u'host1',
                           u'location': u'whq',
                           u'machine': u'virtual',
                           u'memory': 512,
                           u'name': u'test22',
                           u'online': False,
                           u'production': False,
                           u'profile': u'generic',
                           u'rbd_pool': u'rbd.ssd',
                           u'resource_group': u'test',
                           u'resource_group_parent': u'',
                           u'reverses': {},
                           u'service_description': u'asdf',
                           u'servicing': True,
                           u'swap_size': 1073741824,
                           u'timezone': u'Europe/Berlin',
                           u'tmp_size': 5368709120}}
    test22 = json.dumps(cfg).encode('base64')
    test22 = test22.replace('\n', '')

    stdin = StringIO(
        '[{"ModifyIndex": 128, "Value": "%s", "Key": "node/test22"}]' % test22)
    Agent.handle_consul_event(stdin)
    assert util.log_data == [
        'count=1 event=start-consul-events',
        'consul_event=node event=processing-consul-event machine=test22',
        'event=launch-ensure machine=test22',
        'event=finish-consul-events']

    # Applying the same config again doesn't cause another ensure run.
    stdin.seek(0)
    util.log_data = []
    Agent.handle_consul_event(stdin)

    assert util.log_data == [
        'count=1 event=start-consul-events',
        'consul_event=node event=processing-consul-event machine=test22',
        'event=ignore-consul-event machine=test22 reason=config is unchanged',
        'event=finish-consul-events']

    # Changing the config does cause ensure to be called.
    util.log_data = []
    cfg['disk'] = 20
    test22 = json.dumps(cfg).encode('base64')
    test22 = test22.replace('\n', '')

    stdin = StringIO(
        '[{"ModifyIndex": 135, "Value": "%s", "Key": "node/test22"}]' % test22)
    Agent.handle_consul_event(stdin)
    assert util.log_data == ['count=1 event=start-consul-events',
        'consul_event=node event=processing-consul-event machine=test22',
        'event=launch-ensure machine=test22',
        'event=finish-consul-events']


def test_qemu_config_change_physical():
    util.test_log_options['show_events'] = ['consul']

    test22 = json.dumps(
        {u'classes': [u'role::appserver',
                      u'role::backupclient',
                      u'role::generic',
                      u'role::postgresql90',
                      u'role::pspdf',
                      u'role::webproxy'],
         u'name': u'test22',
         u'parameters': {u'ceph_id': u'admin',
                         u'cores': 1,
                         u'directory_password': u'1jmGYd3dddyjpFlLqg63RHie',
                         u'directory_ring': 1,
                         u'disk': 15,
                         u'environment': u'staging',
                         u'id': 4097,
                         u'interfaces': {},
                         u'kvm_host': u'host1',
                         u'location': u'whq',
                         u'machine': u'physical',
                         u'memory': 512,
                         u'name': u'test22',
                         u'online': False,
                         u'production': False,
                         u'profile': u'generic',
                         u'rbd_pool': u'rbd.ssd',
                         u'resource_group': u'test',
                         u'resource_group_parent': u'',
                         u'reverses': {},
                         u'service_description': u'asdf',
                         u'servicing': True,
                         u'swap_size': 1073741824,
                         u'timezone': u'Europe/Berlin',
                         u'tmp_size': 5368709120}}).encode('base64')
    test22 = test22.replace('\n', '')

    stdin = StringIO(
        '[{"ModifyIndex": 123, "Value": "%s", "Key": "node/test22"}]' % test22)
    Agent.handle_consul_event(stdin)
    assert util.log_data == [
        'count=1 event=start-consul-events',
        'event=ignore-consul-event machine=test22 reason=is a physical '
        'machine',
        'event=finish-consul-events']


def test_snapshot_online_vm(vm):
    util.test_log_options['show_events'] = [
        'consul', 'snapshot', 'thaw', 'freeze', 'retry', 'fail']

    vm.ensure_online_local()
    vm.qemu.qmp.close()
    get_log()

    snapshot = json.dumps({'vm': 'simplevm', 'snapshot': 'backy-1234'})
    stdin = StringIO(
        '[{"ModifyIndex": 123, "Value": "%s", '
        '"Key": "snapshot/7468743"}]' % snapshot.encode('base64').strip())
    Agent.handle_consul_event(stdin)
    assert Ellipsis("""\
count=1 event=start-consul-events
event=snapshot machine=simplevm snapshot=backy-1234
event=snapshot-create machine=simplevm name=backy-1234
event=freeze machine=simplevm volume=root
action=continue event=freeze-failed machine=simplevm reason=Unable to sync \
with guest agent after 20 tries.
event=snapshot-ignore machine=simplevm reason=not frozen
event=thaw machine=simplevm volume=root
action=retry event=thaw-failed machine=simplevm reason=Unable to sync with \
guest agent after 20 tries.
action=continue event=thaw-failed machine=simplevm reason=Unable to sync with \
guest agent after 20 tries.
event=handle-key-failed exc_info=True key=snapshot/7468743
event=finish-consul-events""") == get_log()


def test_snapshot_nonexisting_vm():
    util.test_log_options['show_events'] = [
        'consul', 'unknown', 'snapshot']

    get_log()

    snapshot = json.dumps({'vm': 'test77', 'snapshot': 'backy-1234'})
    snapshot = snapshot.encode('base64').replace('\n', '')

    stdin = StringIO(
        '[{"ModifyIndex": 123, "Value": "%s", '
        '"Key": "snapshot/7468743"}]' % snapshot)
    Agent.handle_consul_event(stdin)
    assert get_log() == """\
count=1 event=start-consul-events
event=unknown-vm machine=test77
event=snapshot-ignore machine=test77 reason=failed loading config \
snapshot=backy-1234
event=finish-consul-events"""


def test_snapshot_offline_vm(vm):
    util.test_log_options['show_events'] = [
        'consul', 'snapshot']

    vm.enc['parameters']['kvm_host'] = 'foobar'
    vm.stage_new_config()
    vm.activate_new_config()

    vm.ceph.ensure_root_volume()
    vm.ensure_offline()
    get_log()

    snapshot = json.dumps({'vm': 'simplevm', 'snapshot': 'backy-1234'})
    snapshot = snapshot.encode('base64').replace('\n', '')

    stdin = StringIO(
        '[{"ModifyIndex": 123, "Value": "%s", '
        '"Key": "snapshot/7468743"}]' % snapshot)
    Agent.handle_consul_event(stdin)
    assert get_log() == """\
count=1 event=start-consul-events
event=snapshot machine=simplevm snapshot=backy-1234
event=snapshot-create machine=simplevm name=backy-1234
event=snapshot-ignore machine=simplevm reason=not frozen
event=finish-consul-events"""


def test_multiple_events():
    util.test_log_options['show_events'] = [
        'consul', 'handle-key']

    test22 = json.dumps(
        {u'classes': [u'role::appserver',
                      u'role::backupclient',
                      u'role::generic',
                      u'role::postgresql90',
                      u'role::pspdf',
                      u'role::webproxy'],
         u'name': u'test22',
         u'parameters': {u'ceph_id': u'admin',
                         u'cores': 1,
                         u'directory_password': u'1jmGYd3dddyjpFlLqg63RHie',
                         u'directory_ring': 1,
                         u'disk': 15,
                         u'environment': u'staging',
                         u'id': 4097,
                         u'interfaces': {},
                         u'kvm_host': u'host1',
                         u'location': u'whq',
                         u'machine': u'physical',
                         u'memory': 512,
                         u'name': u'test22',
                         u'online': False,
                         u'production': False,
                         u'profile': u'generic',
                         u'rbd_pool': u'rbd.ssd',
                         u'resource_group': u'test',
                         u'resource_group_parent': u'',
                         u'reverses': {},
                         u'service_description': u'asdf',
                         u'servicing': True,
                         u'swap_size': 1073741824,
                         u'timezone': u'Europe/Berlin',
                         u'tmp_size': 5368709120}}).encode('base64')
    test22 = test22.replace('\n', '')

    stdin = StringIO(
        '[{{"ModifyIndex": 123, "Value": "{0}", "Key": "node/test22"}},'
        '{{"ModifyIndex": 123, "Value": "{0}", "Key": "node/test22"}},'
        '{{"ModifyIndex": 123, "Value": "{0}", "Key": "node/test22"}},'
        '{{"ModifyIndex": 123, "Value": "{0}", "Key": "node/test22"}}]'
        .format(test22))
    Agent.handle_consul_event(stdin)
    assert util.log_data == [
        'count=4 event=start-consul-events',
        'event=handle-key key=node/test22',
        'event=ignore-consul-event machine=test22 reason=is a physical machine',
        'event=finish-handle-key key=node/test22',
        'event=handle-key key=node/test22',
        'event=ignore-consul-event machine=test22 reason=is a physical machine',
        'event=finish-handle-key key=node/test22',
        'event=handle-key key=node/test22',
        'event=ignore-consul-event machine=test22 reason=is a physical machine',
        'event=finish-handle-key key=node/test22',
        'event=handle-key key=node/test22',
        'event=ignore-consul-event machine=test22 reason=is a physical machine',
        'event=finish-handle-key key=node/test22',
        'event=finish-consul-events']



from .ceph import Ceph
import pytest


@pytest.yield_fixture
def ceph_inst():
    cfg = {'resource_group': 'test',
           'rbd_pool': 'rbd.hdd',
           'name': 'simplevm', 'disk': 10}
    ceph = Ceph(cfg)
    ceph.CREATE_VM = 'echo {name}'
    ceph.MKFS_XFS = '-q -f -K'
    ceph.__enter__()
    try:
        yield ceph
    finally:
        ceph.__exit__(None, None, None)



from . import util
from .exc import InvalidCommand, VMConfigNotFound, VMStateInconsistent
from .exc import ConfigChanged
from .hazmat.ceph import Ceph
from .hazmat.qemu import Qemu, detect_current_machine_type
from .incoming import IncomingServer
from .outgoing import Outgoing
from .sysconfig import sysconfig
from .timeout import TimeOut
from .util import rewrite, locate_live_service, MiB, GiB, log
from multiprocessing.pool import ThreadPool
import consulate
import contextlib
import copy
import datetime
import distutils.spawn
import fcntl
import glob
import json
import math
import os
import os.path as p
import pkg_resources
import requests
import shutil
import socket
import subprocess
import sys
import time
import yaml


def _handle_consul_event(event):
    handler = ConsulEventHandler()
    handler.handle(event)


OK, WARNING, CRITICAL, UNKNOWN = 0, 1, 2, 3


class ConsulEventHandler(object):

    def handle(self, event):
        """Actual handling of a single Consul event in a
        separate process."""
        try:
            log.debug("handle-key", key=event['Key'])
            prefix = event['Key'].split('/')[0]
            if not event['Value']:
                log.debug("ignore-key", key=event['Key'], reason='empty value')
                return
            getattr(self, prefix)(event)
        except:  # noqa
            # This must be a bare-except as it protects threads and the main
            # loop from dying. It could be that I'm wrong, but I'm leaving this
            # in for good measure.
            log.exception('handle-key-failed',
                          key=event.get('Key', None),
                          exc_info=True)
        log.debug('finish-handle-key', key=event.get('Key', None))

    def node(self, event):
        config = json.loads(event['Value'].decode('base64'))
        config['consul-generation'] = event['ModifyIndex']
        vm = config['name']
        if config['parameters']['machine'] != 'virtual':
            log.debug('ignore-consul-event',
                      machine=vm, reason='is a physical machine')
            return
        agent = Agent(vm, config)
        log.info('processing-consul-event', consul_event='node', machine=vm)
        if agent.stage_new_config():
            log.info('launch-ensure', machine=vm)
            subprocess.Popen([sys.argv[0], '-D', 'ensure', vm], close_fds=True)
        else:
            log.info('ignore-consul-event',
                     machine=vm, reason='config is unchanged')

    def snapshot(self, event):
        value = json.loads(event['Value'].decode('base64'))
        vm = value['vm']
        snapshot = value['snapshot'].encode('ascii')
        log_ = log.bind(snapshot=snapshot, machine=vm)
        try:
            agent = Agent(vm)
        except RuntimeError:
            log_.warning('snapshot-ignore', reason='failed loading config')
            return
        with agent:
            log_.info('snapshot')
            agent.snapshot(snapshot, keep=0)


def running(expected=True):
    def wrap(f):
        def checked(self, *args, **kw):
            if self.qemu.is_running() != expected:
                self.log.error(
                    f.__name__,
                    expected='VM {}running'.format('' if expected else 'not '))
                raise InvalidCommand(
                    'Invalid command: VM must {}be running to use `{}`.'.
                    format('' if expected else 'not ', f.__name__))
            return f(self, *args, **kw)
        return checked
    return wrap


def locked(blocking=True):
    # This is thread-safe *AS LONG* as every thread uses a separate instance
    # of the agent. Using multiple file descriptors will guarantee that the
    # lock can only be held once even within a single process.
    #
    # However, we ensure locking status even for re-entrant / recursive usage.
    # For that we keep a counter how often we successfully acquired the lock
    # and then unlock when we're back to zero.
    def lock_decorator(f):
        def locked_func(self, *args, **kw):
            # New lockfile behaviour: lock with a global file that is really
            # only used for this purpose and is never replaced.
            self.log.debug('acquire-lock', target=self.lockfile)
            if not self._lockfile_fd:
                if not os.path.exists(self.lockfile):
                    open(self.lockfile, 'a+').close()
                self._lockfile_fd = os.open(self.lockfile, os.O_RDONLY)
            mode = fcntl.LOCK_EX | (fcntl.LOCK_NB if not blocking else 0)
            try:
                fcntl.flock(self._lockfile_fd, mode)
            except IOError:
                # This happens in nonblocking mode and we just give up as
                # that's what's expected to speed up things.
                self.log.info('acquire-lock', result='failed', action='exit',
                              mode='nonblocking')
                return os.EX_TEMPFAIL
            self.log.debug('acquire-lock',
                           target=self.lockfile, result='locked')

            self._lock_count += 1
            self.log.debug('lock-status', count=self._lock_count)
            try:
                return f(self, *args, **kw)
            finally:
                self._lock_count -= 1
                self.log.debug('lock-status', count=self._lock_count)

                if self._lock_count == 0:
                    # New
                    try:
                        self.log.debug('release-lock', target=self.lockfile)
                        fcntl.flock(self._lockfile_fd, fcntl.LOCK_UN)
                        self.log.debug('release-lock', target=self.lockfile,
                                       result='unlocked')
                    except:  # noqa
                        self.log.debug('release-lock', exc_info=True,
                                       target=self.lockfile, result='error')
                        pass
        return locked_func
    return lock_decorator


def swap_size(memory):
    """Returns the swap partition size in bytes."""
    swap_mib = max(1024, 32 * math.sqrt(memory))
    return int(swap_mib) * MiB


def tmp_size(disk):
    """Returns the tmp partition size in bytes."""
    tmp_gib = max(5, math.sqrt(disk))
    return int(tmp_gib) * GiB


class Agent(object):
    """The agent to control a single VM."""

    # Attributes on this class can be overriden (in a controlled fashion
    # from the sysconfig module. See this class' __init__. The defaults
    # are here to support testing.

    this_host = ''
    migration_ctl_address = None
    accelerator = ''
    machine_type = 'pc-i440fx'
    vhost = False
    ceph_id = 'admin'
    timeout_graceful = 10
    vm_config_template_path = [
        '/etc/qemu/qemu.vm.cfg.in',
        pkg_resources.resource_filename(__name__, 'qemu.vm.cfg.in')
    ]
    consul_token = None
    consul_generation = -1

    # The binary generation is used to signal into a VM that the host
    # environment has changed in a way that requires a _cold_ reboot, thus
    # asking the VM to shut down cleanly so we can start it again. Change this
    # introducing a new Qemu machine type, or providing security fixes that
    # can't be applied by live migration.
    # The counter should be increased through the platform management as
    # this may change independently from fc.qemu releases.
    binary_generation = 0

    # For upgrade-purposes we're running an old and a new locking mechanism
    # in step-lock. We used to lock the configfile but we're using rename to
    # update it atomically. That's not compatible and will result in a consul
    # event replacing the file and then getting a lock on the new file while
    # there still is another process running. This will result in problems
    # accessing the QMP socket as that only accepts a single connection and
    # will then time out.
    _lock_count = 0
    _lockfile_fd = None
    _configfile_fd = None

    def __init__(self, name, enc=None):
        # Update configuration values from system or test config.
        self.log = log.bind(machine=name)

        self.__dict__.update(sysconfig.agent)

        self.name = name

        # The ENC data is the parsed configuration from the config file, or
        # from consul (or from wherever). This is intended to stay true to
        # the config file format (not adding computed values or manipulating
        # them) so that we can generate a new config file based on this data
        # through `stage_new_config()`.
        if enc is not None:
            self.enc = enc
        else:
            self.enc = self._load_enc()

        self.consul = consulate.Consul(token=self.consul_token)
        self.contexts = ()

    @property
    def configfile(self):
        return '/etc/qemu/vm/{}.cfg'.format(self.name)

    @property
    def configfile_staging(self):
        return '/etc/qemu/vm/.{}.cfg.staging'.format(self.name)

    @property
    def lockfile(self):
        return '/run/qemu.{}.lock'.format(self.name)

    def _load_enc(self):
        try:
            with open(self.configfile) as f:
                return yaml.safe_load(f)
        except IOError:
            if os.path.exists(self.configfile_staging):
                # The VM has been freshly created. Set up a boilerplate
                # configuration and le the actual config be established
                # through the 'ensure' command.
                return None
            else:
                self.log.error('unknown-vm')
                raise VMConfigNotFound("Could not load {}".format(
                    self.configfile))

    @classmethod
    def handle_consul_event(cls, input=sys.stdin):
        events = json.load(input)
        if not events:
            return
        log.info('start-consul-events', count=len(events))
        pool = ThreadPool(sysconfig.agent.get('consul_event_threads', 3))
        for event in sorted(events, key=lambda e: e.get('Key')):
            pool.apply_async(_handle_consul_event, (event,))
            time.sleep(0.05)
        pool.close()
        pool.join()
        log.info('finish-consul-events')

    @classmethod
    def _vm_agents_for_host(cls):
        for candidate in sorted(glob.glob('/run/qemu.*.pid')):
            name = candidate.replace('/run/qemu.', '')
            name = name.replace('.pid', '')
            try:
                agent = Agent(name)
                yield agent
            except Exception:
                log.exception('load-agent', machine=name, exc_info=True)

    @classmethod
    def ls(cls):
        """List all VMs that this host knows about.

        This means specifically that we have status (PID) files for the
        VM in the expected places.

        Gives a quick status for each VM.
        """
        for vm in cls._vm_agents_for_host():
            with vm:
                running = vm.qemu.process_exists()

                if running:
                    vm_mem = vm.qemu.proc().memory_full_info()

                    expected_size = (
                        vm.cfg['memory'] * 1024 * 1024 +
                        vm.qemu.vm_expected_overhead * 1024 * 1024)

                    log.info('online',
                             machine=vm.name,
                             cores=vm.cfg['cores'],
                             memory_booked='{:,.0f}'.format(vm.cfg['memory']),
                             memory_pss='{:,.0f}'.format(vm_mem.pss / MiB),
                             memory_swap='{:,.0f}'.format(vm_mem.swap / MiB))

                else:
                    log.info('offline', machine=vm.name)

    @classmethod
    def check(cls):
        """Perform a health check of this host from a Qemu perspective.

        Checks:

        * no VM exceeds their PSS by guest memory + 2x expected overhead
        * VMs have very little swap (<1 GiB)
        * the total of the VMs PSS does not exceed
          total of guest + expected overhead

        Sets exit code according to the Nagios specification.

        """
        vms = list(cls._vm_agents_for_host())

        large_overhead_vms = []
        swapping_vms = []
        total_guest_and_overhead = 0
        expected_guest_and_overhead = 0

        # individual VMs ok?
        for vm in vms:
            with vm:
                try:
                    vm_mem = vm.qemu.proc().memory_full_info()
                except Exception:
                    # It's likely that the process went away while we analyzed
                    # it. Ignore.
                    continue
                if vm_mem.swap > 1 * GiB:
                    swapping_vms.append(vm)
                expected_size = (
                    vm.cfg['memory'] * MiB +
                    2 * vm.qemu.vm_expected_overhead * MiB)
                expected_guest_and_overhead += (
                    vm.cfg['memory'] * MiB +
                    vm.qemu.vm_expected_overhead * MiB)
                total_guest_and_overhead += vm_mem.pss
                if vm_mem.pss > expected_size:
                    large_overhead_vms.append(vm)

        output = []
        result = OK
        if large_overhead_vms:
            result = WARNING
            output.append('VMs with large overhead: ' +
                          ','.join(x.name for x in large_overhead_vms))
        if swapping_vms:
            result = WARNING
            output.append('VMs swapping:' +
                          ','.join(x.name for x in swapping_vms))
        if total_guest_and_overhead > expected_guest_and_overhead:
            result = CRITICAL
            output.append('High total overhead')

        if result is OK:
            output.insert(0, 'OK')
        elif result is WARNING:
            output.insert(0, 'WARNING')
        elif result is CRITICAL:
            output.insert(0, 'CRITICAL')
        else:
            output.insert(0, 'UNKNOWN')

        output.insert(
            1, '{} VMs'.format(len(vms)))
        output.insert(
            2, '{:,.0f} MiB used'.format(total_guest_and_overhead / MiB))
        output.insert(
            3, '{:,.0f} MiB expected'.format(expected_guest_and_overhead / MiB))

        print(' - '.join(output))

        return result

    def stage_new_config(self):
        """Save the current config on the agent into a staging config file.

        This method is safe to call outside of the agent's context manager.

        """
        # Ensure the config directory exists
        if not p.isdir(p.dirname(self.configfile)):
            os.makedirs(p.dirname(self.configfile))

        # Ensure the staging file exists.
        open(self.configfile_staging, 'a+').close()
        staging_lock = os.open(self.configfile_staging, os.O_RDONLY)
        fcntl.flock(staging_lock, fcntl.LOCK_EX)
        self.log.debug(
            'acquire-staging-lock', target=self.configfile_staging,
            result='locked')
        try:
            # Verify generation of config to protect against lost updates.
            with open(self.configfile_staging, 'r') as current_staging:
                try:
                    current_staging_config = yaml.safe_load(current_staging)
                    current_staging_generation = current_staging_config[
                        'consul-generation']
                except Exception:
                    self.log.debug('inconsistent-staging-config')
                    pass
                else:
                    if (self.enc['consul-generation'] <=
                            current_staging_generation):
                        # Stop right here, do not write a new config if the
                        # existing one is newer (or as new) already.
                        self.log.debug('ignore-old-update',
                                       existing=self.consul_generation,
                                       update=current_staging_generation)
                        return False

            # The config is either newer, or there is no staging config, or it
            # is inconsistent. Let's update it.
            self.log.debug('save-staging-config')
            # Update the file in place to avoid lock breaking.
            with open(self.configfile_staging, 'w') as new_staging:
                yaml.safe_dump(self.enc, new_staging)
            os.chmod(self.configfile_staging, 0o644)
            return True
        finally:
            fcntl.flock(staging_lock, fcntl.LOCK_UN)
            os.close(staging_lock)
            self.log.debug(
                'release-staging-lock',
                target=self.configfile_staging, result='released')

    @locked()
    def activate_new_config(self):
        """Activate the current staged config (if any).

        After calling this method, the agent's context manager must be
        re-entered to activate the changed configuration.

        """
        if not os.path.exists(self.configfile_staging):
            self.log.debug('check-staging-config', result='none')
            return False

        staging_lock = os.open(self.configfile_staging, os.O_RDONLY)
        fcntl.flock(staging_lock, fcntl.LOCK_EX)
        self.log.debug(
            'acquire-staging-lock', target=self.configfile_staging,
            result='locked')
        try:
            # Verify generation of config to protect against lost updates.
            with open(self.configfile_staging, 'r') as current_staging:
                try:
                    current_staging_config = yaml.safe_load(current_staging)
                    staging_generation = current_staging_config[
                        'consul-generation']
                except Exception:
                    self.log.debug('update-check',
                                   result='inconsistent', action='purge')
                    os.unlink(self.configfile_staging)
                    return False
                else:
                    if staging_generation <= self.consul_generation:
                        # Stop right here, do not write a new config if the
                        # existing one is newer (or as new) already.
                        self.log.debug(
                            'update-check',
                            result='stale-update', action='ignore',
                            update=staging_generation,
                            current=self.consul_generation)
                        # The old staging file needs to stay around so that
                        # the consul writer knows whether to launch an ensure
                        # agent or not.
                        return False
                self.log.debug('update-check',
                               result='update-available',
                               action='update',
                               update=staging_generation,
                               current=self.consul_generation)

            # The config seems consistent and newer, lets update.
            # We can replace the config file because that one is protected
            # by the global VM lock.
            shutil.copy2(self.configfile_staging, self.configfile)
            self.enc = self._load_enc()
            return True
        finally:
            fcntl.flock(staging_lock, fcntl.LOCK_UN)
            os.close(staging_lock)
            self.log.debug(
                'release-staging-lock', target=self.configfile_staging,
                result='released')

    def has_new_config(self):
        if not os.path.exists(self.configfile_staging):
            self.log.debug('check-staging-config', result='none')
            return False

        staging_lock = os.open(self.configfile_staging, os.O_RDONLY)
        fcntl.flock(staging_lock, fcntl.LOCK_EX)
        self.log.debug(
            'acquire-staging-lock', target=self.configfile_staging,
            result='locked')
        try:
            # Verify generation of config to protect against lost updates.
            with open(self.configfile_staging, 'r') as current_staging:
                try:
                    current_staging_config = yaml.safe_load(current_staging)
                    staging_generation = current_staging_config[
                        'consul-generation']
                except Exception:
                    self.log.debug('update-check',
                                   result='inconsistent', action='purge')
                    os.unlink(self.configfile_staging)
                    return False
                else:
                    if staging_generation <= self.consul_generation:
                        # Stop right here, do not write a new config if the
                        # existing one is newer (or as new) already.
                        self.log.debug('update-check',
                                       result='stale-update',
                                       action='ignore',
                                       update=staging_generation,
                                       current=self.consul_generation)
                        # The old staging file needs to stay around so that
                        # the consul writer knows whether to launch an ensure
                        # agent or not.
                        return False
                self.log.debug('update-check',
                               result='update-available',
                               action='update',
                               update=staging_generation,
                               current=self.consul_generation)
            return True
        finally:
            fcntl.flock(staging_lock, fcntl.LOCK_UN)
            os.close(staging_lock)
            self.log.debug(
                'release-staging-lock', target=self.configfile_staging,
                result='released')

    def __enter__(self):
        # Allow updating our config by exiting/entering after setting new ENC
        # data.
        # This copy means we can't manipulate cfg to update ENC data, which is
        # OK. We did this at some point and we're adding computed data to the
        # `cfg` structure, that we do not want to accidentally reflect back
        # into the config file.
        if self.enc is None:
            return
        self.cfg = copy.copy(self.enc['parameters'])
        self.cfg['name'] = self.enc['name']
        self.cfg['swap_size'] = swap_size(self.cfg['memory'])
        self.cfg['tmp_size'] = tmp_size(self.cfg['disk'])
        self.cfg['ceph_id'] = self.ceph_id
        try:
            self.consul_generation = self.enc['consul-generation']
        except Exception:
            import pdb; pdb.set_trace()
        self.qemu = Qemu(self.cfg)
        self.ceph = Ceph(self.cfg)
        self.contexts = [self.qemu, self.ceph]
        for attr in ['migration_ctl_address']:
            setattr(self, attr, getattr(self, attr).format(**self.cfg))
        for cand in self.vm_config_template_path:
            if p.exists(cand):
                self.vm_config_template = cand
                break

        for c in self.contexts:
            c.__enter__()

    def __exit__(self, exc_value, exc_type, exc_tb):
        for c in self.contexts:
            try:
                c.__exit__(exc_value, exc_type, exc_tb)
            except Exception:
                self.log.exception('leave-subsystems', exc_info=True)

    def ensure(self):
        # Run the config activation code at least once, but then only if there
        # are still config changes around. We do this because multiple updates
        # may be piling up and we want to catch up as quickly as possible but
        # we only stage the updates and don't want to spawn an agent for each
        # update of each VM. So if an agent is already running it checks again
        # after it did something and then keeps going. OTOH agents that are
        # spawned after staging a new config will give up immediately when the
        # notice another agent running.

        # The __exit__/__enter__ dance is to simplify handling this special
        # case where we have to handle this method differently as it keeps
        # entering/exiting while cycling through the config changes. This keeps
        # the API and unit tests a bit cleaner even though we enter/exit one
        # time too many.
        self.__exit__(None, None, None)
        try:
            first = True
            while self.has_new_config() or first:
                self.log.info('running-ensure',
                              generation=self.consul_generation)
                first = False
                try:
                    locking_code = self.ensure_()
                    if locking_code == os.EX_TEMPFAIL:
                        # We didn't get the lock, so someone else is
                        # already around who will pick up the (additional)
                        # changed config later.
                        break
                except ConfigChanged:
                    # Well then. Let's try this again.
                    continue
        finally:
            self.__enter__()
        self.log.debug('changes-settled')

    @locked(blocking=False)
    def ensure_(self):
        self.activate_new_config()
        with self:
            # Host assignment is a bit tricky: we decided to not interpret an
            # *empty* cfg['kvm_host'] as "should not be running here" for the
            # sake of not accidentally causing downtime.
            try:
                if not self.cfg['online']:
                    # Wanted offline.
                    self.ensure_offline()

                elif self.cfg['kvm_host']:
                    if self.cfg['kvm_host'] != self.this_host:
                        # Wanted online, but explicitly on a different host.
                        self.ensure_online_remote()
                    else:
                        # Wanted online, and it's OK if it's running here, but
                        # I'll only start it if it really is wanted here.
                        self.ensure_online_local()

                self.raise_if_inconsistent()
            except VMStateInconsistent:
                # Last-resort seat-belt to verify that we ended up in a
                # consistent state. Inconsistent states result in the VM being
                # forcefully terminated.
                self.log.error('inconsistent-state', action='destroy',
                               exc_info=True)
                self.qemu.destroy()

    def ensure_offline(self):
        if self.qemu.is_running():
            self.log.info(
                'ensure-state', wanted='offline', found='online',
                action='stop')
            self.stop()
        else:
            self.log.info(
                'ensure-state', wanted='offline', found='offline',
                action='none')

    def ensure_online_remote(self):
        if not self.qemu.is_running():
            # This cleans up potential left-overs from a previous migration.
            self.ceph.stop()
            self.consul_deregister()
            self.cleanup()
            return

        migration_errors = self.outmigrate()
        if migration_errors:
            # Stop here, do not clean up. This is fine as we may not have
            # made contact with our destination host. We keep trying later.
            return

        # Ensure ongoing maintenance for a VM that was potentially outmigrated.
        # But where the outmigration may have left bits and pieces.
        self.ceph.stop()
        self.consul_deregister()
        self.cleanup()

    def ensure_online_local(self):
        # Ensure state
        agent_likely_ready = True
        if not self.qemu.is_running():
            current_host = self._requires_inmigrate_from()
            if current_host:
                self.log.info(
                    'ensure-state', wanted='online', found='offline',
                    action='inmigrate', remote=current_host)
                exitcode = self.inmigrate()
                if exitcode:
                    # This is suboptimal: I hate error returns,
                    # but the main method is also a command. If we did
                    # not succeed in migrating, then I also don't want the
                    # consul registration to happen.
                    return
            else:
                self.log.info(
                    'ensure-state', wanted='online', found='offline',
                    action='start')
                self.start()
                agent_likely_ready = False
        else:
            self.log.info(
                'ensure-state', wanted='online', found='online', action='')

        # Perform ongoing adjustments of the operational parameters of the
        # running VM.
        self.consul_register()
        self.ensure_online_disk_size()
        self.ensure_online_disk_throttle()
        self.ensure_watchdog()
        # Be aggressive/opportunistic about re-acquiring locks in case
        # they were taken away.
        self.ceph.lock()
        if agent_likely_ready:
            # This requires guest agent interaction and we should only
            # perform this when we haven't recently booted the machine to
            # reduce the time we're unnecessarily waiting for timeouts.
            self.ensure_thawed()
            self.mark_qemu_binary_generation()

    def cleanup(self):
        """Removes various run and tmp files."""
        self.qemu.clean_run_files()
        for tmp in glob.glob(self.configfile + '?*'):
            os.unlink(tmp)

    def ensure_thawed(self):
        self.log.info('ensure-thawed', volume='root')
        try:
            self.qemu.thaw()
        except Exception as e:
            self.log.error('ensure-thawed-failed', reason=str(e))

    def mark_qemu_binary_generation(self):
        self.log.info(
            'mark-qemu-binary-generation', generation=self.binary_generation)
        try:
            self.qemu.write_file('/run/qemu-binary-generation-current',
                                 str(self.binary_generation) + '\n')
        except Exception as e:
            self.log.error('mark-qemu-binary-generation', reason=str(e))

    def ensure_online_disk_size(self):
        """Trigger block resize action for the root disk."""
        target_size = self.cfg['disk'] * (1024 ** 3)
        if self.ceph.root.size >= target_size:
            self.log.info('check-disk-size',
                          wanted=target_size,
                          found=self.ceph.root.size,
                          action='none')
            return
        self.log.info('check-disk-size', wanted=target_size,
                      found=self.ceph.root.size, action='resize')
        self.qemu.resize_root(target_size)

    def ensure_online_disk_throttle(self):
        """Ensure throttling settings."""
        target = self.cfg.get(
            'iops',
            self.qemu.throttle_by_pool.get(self.cfg['rbd_pool'], 250))
        devices = self.qemu.block_info()
        for device in devices.values():
            current = device['inserted']['iops']
            if current != target:
                self.log.info('ensure-throttle', device=device['device'],
                              target_iops=target, current_iops=current,
                              action='throttle')
                self.qemu.block_io_throttle(device['device'], target)
            else:
                self.log.info('ensure-throttle', device=device['device'],
                              target_iops=target, current_iops=current,
                              action='none')

    def ensure_watchdog(self, action='none'):
        """Ensure watchdog settings."""
        self.log.info('ensure-watchdog', action=action)
        self.qemu.watchdog_action(action)

    @property
    def svc_name(self):
        """Consul service name."""
        return 'qemu-{}'.format(self.name)

    def consul_register(self):
        """Register running VM with Consul."""
        self.log.debug('consul-register')
        self.consul.agent.service.register(
            self.svc_name,
            address=self.this_host,
            interval='5s',
            check=('test -e /proc/$(< /run/qemu.{}.pid )/mem || exit 2'.
                   format(self.name)))

    def consul_deregister(self):
        """De-register non-running VM with Consul."""
        try:
            if self.svc_name not in self.consul.agent.services()[0]:
                return
            self.log.info('consul-deregister')
            self.consul.agent.service.deregister('qemu-{}'.format(self.name))
        except requests.exceptions.ConnectionError:
            pass
        except Exception:
            self.log.exception('consul-deregister-failed', exc_info=True)

    @locked()
    @running(False)
    def start(self):
        self.generate_config()
        self.ceph.start(self.enc, self.binary_generation)
        self.qemu.start()
        self.consul_register()
        self.ensure_online_disk_throttle()
        self.ensure_watchdog()
        # We exit here without releasing the ceph lock in error cases
        # because the start may have failed because of an already running
        # process. Removing the lock in that case is dangerous. OTOH leaving
        # the lock is never dangerous. Lets go with that.

    @contextlib.contextmanager
    def frozen_vm(self):
        """Ensure a VM has a non-changing filesystem, that is clean enough
        to be mounted with a straight-forward journal replay.

        This is implemented using the 'fs-freeze' API through the Qemu guest
        agent.

        As we can't indicate the cleanliness of a non-running
        (potentially crashed) VM we do only signal a successful freeze iff
        the freeze happened and succeeded.

        """
        frozen = False
        try:
            if self.qemu.is_running():
                self.log.info('freeze', volume='root')
                try:
                    self.qemu.freeze()
                    frozen = True
                except Exception as e:
                    self.log.error('freeze-failed',
                                   reason=str(e), action='continue',
                                   machine=self.name)
            yield frozen
        finally:
            if self.qemu.is_running():
                try:
                    self.log.info('thaw', volume='root')
                    self.qemu.thaw()
                except Exception as e:
                    self.log.error('thaw-failed', reason=str(e),
                                   action='retry')
                    try:
                        self.qemu.thaw()
                    except Exception as e:
                        self.log.error('thaw-failed', reason=str(e),
                                       action='continue')
                        raise

    @locked()
    def snapshot(self, snapshot, keep=0):
        """Guarantees a _consistent_ snapshot to be created.

        If we can't properly freeze the VM then whoever needs a (consistent)
        snapshot needs to figure out whether to go forward with an
        inconsistent snapshot.
        """
        if keep:
            until = util.today() + datetime.timedelta(days=keep)
            snapshot = snapshot + '-keep-until-' + until.strftime('%Y%m%d')
        if snapshot in [x.snapname for x in self.ceph.root.snapshots]:
            self.log.info('snapshot-exists', snapshot=snapshot)
            return
        self.log.debug('snapshot-create', name=snapshot)
        with self.frozen_vm() as frozen:
            if frozen:
                self.ceph.root.snapshots.create(snapshot)
            else:
                self.log.debug('snapshot-ignore', reason='not frozen')

    @locked()
    def status(self):
        """Determine status of the VM."""
        if self.qemu.is_running():
            status = 0
            self.log.info('vm-status', result='online')
            for device in self.qemu.block_info().values():
                self.log.info('disk-throttle',
                              device=device['device'],
                              iops=device['inserted']['iops'])
        else:
            status = 1
            self.log.info('vm-status', result='offline')
        for volume in self.ceph.volumes:
            locker = volume.lock_status()
            self.log.info('rbd-status', volume=volume.fullname, locker=locker)
        consul = locate_live_service(self.consul, 'qemu-' + self.name)
        if consul:
            self.log.info('consul', service=consul['Service'], address=consul['Address'])
        else:
            self.log.info('consul', result='no response')
        return status

    def telnet(self):
        """Open telnet connection to the VM monitor."""
        self.log.info('connect-via-telnet')
        telnet = distutils.spawn.find_executable('telnet')
        os.execv(telnet, ('telnet', 'localhost', str(self.qemu.monitor_port)))

    @locked()
    @running(True)
    def stop(self):
        timeout = TimeOut(self.timeout_graceful, interval=3)
        self.log.info('graceful-shutdown')
        try:
            self.qemu.graceful_shutdown()
        except (socket.error, RuntimeError):
            pass
        while timeout.tick():
            self.log.debug('checking-offline', remaining=timeout.remaining)
            if not self.qemu.is_running():
                self.log.info('vm-offline')
                self.ceph.stop()
                self.consul_deregister()
                self.cleanup()
                self.log.info('graceful-shutdown-completed')
                break
        else:
            self.log.warn('graceful-shutdown-failed', reason='timeout')
            self.kill()

    @locked()
    def restart(self):
        self.log.info('restart-vm')
        self.stop()
        self.start()

    @locked()
    @running(True)
    def kill(self):
        self.log.info('kill-vm')
        timeout = TimeOut(15, interval=1, raise_on_timeout=True)
        self.qemu.destroy()
        while timeout.tick():
            if not self.qemu.is_running():
                self.log.info('killed-vm')
                self.ceph.stop()
                self.consul_deregister()
                self.cleanup()
                break
        else:
            self.log.warning('kill-vm-failed', note='Check lock consistency.')

    def _requires_inmigrate_from(self):
        """Check whether an inmigration makes sense.

        This makes sense if the VM isn't running locally and (Consul knows
        about a running VM or the VM is locked remotely).

        Returns the name of the remote host the VM should be migrated from
        or None if inmigration doesn't make sense.

        """
        existing = locate_live_service(self.consul, 'qemu-' + self.name)

        if existing and existing['Address'] != self.this_host:
            # Consul knows about a running VM. Lets try a migration.
            return existing['Address']

        if not self.ceph.is_unlocked():
            # Consul doesn't know about a running VM and no volume is locked.
            # It doesn't make sense to live migrate this VM.
            return None

        if self.ceph.locked_by_me():
            # Consul doesn't know about a running VM and the volume is
            # locked by me, so it doesn't make sense to live migrate the VM.
            return None

        # The VM seems to be locked somewhere else, try to migrate it from
        # there.
        return self.ceph.locked_by()

    @locked()
    @running(False)
    def inmigrate(self):
        self.log.info('inmigrate')
        server = IncomingServer(self)
        exitcode = server.run()
        if not exitcode:
            self.consul_register()
        self.log.info('inmigrate-finished', exitcode=exitcode)
        return exitcode

    @locked()
    @running(True)
    def outmigrate(self):
        self.log.info('outmigrate')
        # re-register in case services got lost during Consul restart
        self.consul_register()
        client = Outgoing(self)
        exitcode = client()
        if not exitcode:
            self.consul_deregister()
        self.log.info('outmigrate-finished', exitcode=exitcode)
        return exitcode

    @locked()
    def lock(self):
        self.log.info('assume-all-locks')
        for vol in self.ceph.volumes:
            vol.lock()

    @locked()
    @running(False)
    def unlock(self):
        self.log.info('release-all-locks')
        self.ceph.stop()

    @running(False)
    def force_unlock(self):
        self.log.info('break-all-locks')
        self.ceph.force_unlock()

    # Helper methods
    def raise_if_inconsistent(self):
        """Raise an VMStateInconsistent error if the VM state is not
        consistent.

        Either all of the following or none of the following must be true:

        - The process belonging to the PID we know exists.
        - The QMP monitor reliably tells us that the Qemu status of the
          VM/CPUs is "running"
        - We have locked the volumes in Ceph.

        If they are not the same, then the state is considered inconsistent
        and this method will return False.

        """
        state = VMStateInconsistent()
        state.qemu = self.qemu.is_running()
        state.proc = bool(self.qemu.proc())
        state.ceph_lock = self.ceph.locked_by_me()
        self.log.debug('check-state-consistency',
                       is_consistent=state.is_consistent(),
                       qemu=state.qemu,
                       proc=state.proc,
                       ceph_lock=state.ceph_lock)
        if not state.is_consistent():
            raise state

    # CAREFUL: changing anything in this config files will require maintenance
    # w/ reboot of all VMs in the infrastructure. Need to increase the
    # binary generation for that, though.
    def generate_config(self):
        """Generate a new Qemu config (and options) for a freshly
        starting VM.

        The configs are intended to be non-host-specific.

        This two-step behaviour is needed to support migrating VMs
        and keeping arguments consistent while allowing to localize arguments
        to the host running or receiving the VM.

        """
        self.log.debug('generate-config')
        self.qemu.args = [
            '-nodefaults',
            '-name {name},process=kvm.{name}',
            '-chroot {{chroot}}',
            '-runas nobody',
            '-serial file:/var/log/vm/{name}.log',
            '-display vnc={{vnc}}',
            '-pidfile {{pidfile}}',
            '-vga std',
            # We use this '-m' flag to find what a running VM is actually
            # using at the moment. If this flag is changed then that code must
            # be adapted as well. This is used in incoming.py and qemu.py.
            '-m {memory}',
            '-readconfig {{configfile}}']
        self.qemu.args = [a.format(**self.cfg)
                          for a in self.qemu.args]

        vhost = '  vhost = "on"' if self.vhost else ''

        netconfig = []
        for net, net_config in sorted(self.cfg['interfaces'].items()):
            ifname = 't{}{}'.format(net, self.cfg['id'])
            netconfig.append("""
[device]
  driver = "virtio-net-pci"
  netdev = "{ifname}"
  mac = "{mac}"

[netdev "{ifname}"]
  type = "tap"
  ifname = "{ifname}"
  script = "/etc/kvm/kvm-ifup"
  downscript = "/etc/kvm/kvm-ifdown"
{vhost}
""".format(ifname=ifname, mac=net_config['mac'], vhost=vhost))

        with open(self.vm_config_template) as f:
            tpl = f.read()
        accelerator = ('  accel = "{}"'.format(self.accelerator)
                       if self.accelerator else '')
        machine_type = detect_current_machine_type(self.machine_type)
        self.qemu.config = tpl.format(
            accelerator=accelerator,
            machine_type=machine_type,
            disk_cache_mode=self.qemu.disk_cache_mode,
            network=''.join(netconfig), **self.cfg)



from .agent import Agent
import os
import pkg_resources
import pytest
import shutil
import structlog
import subprocess
import sys
import traceback


def pytest_collectstart(collector):
    # sys.modules['rados'] = mock.Mock()
    # sys.modules['rbd'] = mock.Mock()
    from fc.qemu.sysconfig import sysconfig
    sysconfig.load_system_config()


@pytest.fixture(scope='session')
def setup_structlog():
    from . import util
    log_exceptions = False  # set to True to temporarily get detailed tracebacks

    def test_logger(logger, method_name, event):
        result = []

        show_methods = util.test_log_options['show_methods']
        if show_methods and method_name not in show_methods:
            raise structlog.DropEvent

        show_events = util.test_log_options['show_events']
        if show_events:
            for show in show_events:
                if show in event['event']:
                    break
            else:
                raise structlog.DropEvent

        if log_exceptions:
            stack = event.pop("stack", None)
            exc = event.pop("exception", None)
        for key in sorted(event):
            result.append('{}={}'.format(key, event[key]))
        util.log_data.append(' '.join(result))
        if log_exceptions:
          if stack:
              util.log_data.extend(stack.splitlines())
          if exc:
              util.log_data.extend(exc.splitlines())
        raise structlog.DropEvent

    structlog.configure(processors=(
        ([structlog.processors.format_exc_info] if log_exceptions else []) +
        [test_logger]))


@pytest.fixture(autouse=True)
def reset_structlog(setup_structlog):
    from . import util
    util.log_data = []
    util.test_log_options = {
        'show_methods': [],
        'show_events': []}


def pytest_assertrepr_compare(op, left, right):
    if left.__class__.__name__ == 'Ellipsis':
        return left.compare(right).diff
    elif right.__class__.__name__ == 'Ellipsis':
        return right.compare(left).diff


@pytest.yield_fixture
def clean_environment():
    def clean():
        subprocess.call('pkill -f qemu', shell=True)
        subprocess.call('rbd rm rbd.ssd/simplevm.swap', shell=True)
        subprocess.call('rbd snap purge rbd.ssd/simplevm.root', shell=True)
        subprocess.call('rbd rm rbd.ssd/simplevm.root', shell=True)
        subprocess.call('rbd rm rbd.ssd/simplevm.tmp', shell=True)
    clean()
    yield
    clean()


@pytest.yield_fixture
def vm(clean_environment):
    fixtures = pkg_resources.resource_filename(__name__, 'tests/fixtures')
    shutil.copy(fixtures + '/simplevm.yaml', '/etc/qemu/vm/simplevm.cfg')
    if os.path.exists('/etc/qemu/vm/.simplevm.cfg.staging'):
        os.unlink('/etc/qemu/vm/.simplevm.cfg.staging')
    vm = Agent('simplevm')
    vm.timeout_graceful = 1
    vm.__enter__()
    vm.qemu.guestagent_timeout = .1
    vm.qemu.qmp_timeout = .1
    vm.qemu.vm_expected_overhead = 128
    for snapshot in vm.ceph.root.snapshots:
        snapshot.remove()
    vm.qemu.destroy()
    vm.unlock()
    get_log()
    yield vm

    for snapshot in vm.ceph.root.snapshots:
        snapshot.remove()
    exc_info = sys.exc_info()
    vm.__exit__(*exc_info)
    if len(exc_info):
        print(traceback.print_tb(exc_info[2]))
    os.unlink('/etc/qemu/vm/simplevm.cfg')
    if os.path.exists('/etc/qemu/vm/.simplevm.cfg.staging'):
        os.unlink('/etc/qemu/vm/.simplevm.cfg.staging')


def get_log():
    from fc.qemu import util
    result = '\n'.join(util.log_data)
    util.log_data = []
    return result



from ..timeout import TimeOut, TimeoutError
import pytest


def test_timeout_raises():
    timeout = TimeOut(.1, raise_on_timeout=True)
    with pytest.raises(TimeoutError):
        while timeout.tick():
            pass


def test_timeout_stops():
    timeout = TimeOut(.1)
    while timeout.tick():
        pass



from fc.qemu.main import daemonize
import datetime
import sys
import subprocess
import time
import shlex


def run_supervised(cmd, name, logfile):
    daemonize()
    log = open(logfile, 'a+', buffering=0)
    now = datetime.datetime.now().isoformat()
    log.write('{} - starting command {}\n'.format(now, cmd))
    s = subprocess.Popen(
        shlex.split(cmd), close_fds=True, stdin=None, stdout=log, stderr=log)
    now = datetime.datetime.now().isoformat()
    log.write('{} - command has PID {}\n'.format(now, s.pid))
    exit_code = s.wait()
    now = datetime.datetime.now().isoformat()
    log.write('{} - command exited with exit code {}\n'.format(now, exit_code))


if __name__ == '__main__':
    run_supervised(*sys.argv[1:])



from ..util import log
import fcntl
import json
import random
import socket


class ClientError(RuntimeError):
    pass


class GuestAgent(object):
    """Wraps qemu guest agent wire protocol."""

    def __init__(self, machine, timeout):
        self.machine = machine
        self.timeout = timeout
        self.log = log.bind(machine=machine)
        self.file = None
        self.client = None

    def read(self):
        """Reads single response from the GA and returns the result.

        Blocks and runs into a timeout if no response is available.
        """
        result = json.loads(self.file.readline())
        if 'error' in result:
                raise ClientError(result)
        return result['return']

    def cmd(self, cmd, flush_ga_parser=False, timeout=None, **args):
        """Issues GA command and returns the result."""
        message = json.dumps({"execute": cmd, "arguments": args})
        if flush_ga_parser:
            # \xff is an invalid utf-8 character and recommended to safely
            # ensure that the parser of the guest agent at the other end
            # is reset to a known state. This is recommended for sync.
            # http://wiki.qemu-project.org/index.php/Features/GuestAgent#guest-sync
            message = b'\xff' + message
        timeout = timeout or self.timeout
        # Allow setting temporary timeouts for operations that are known to be
        # slow.
        self.client.settimeout(timeout)
        self.client.send(message)
        return self.read()

    def sync(self):
        """Ensures that request and response are in order."""
        sync_id = random.randint(0, 0xffff)
        n = 0
        try:
            result = self.cmd('guest-sync', id=sync_id, flush_ga_parser=True)
        except ClientError:
            # we tripped a client error as we caused the guest agent to notice
            # invalid json, which in turn triggers an error response
            result = self.read()
        except socket.error:
            # Maybe a timeout. Keep trying a little bit harder.
            result = None
        while n < 20:
            if result == sync_id:
                return
            self.log.error('incorrect-sync-id', expected=sync_id, got=result,
                           tries=n)
            n += 1
            try:
                result = self.read()
            except (ClientError, socket.error):
                # we tripped a client error later than right now. There may
                # have been a response still in the queue.
                pass

        raise ClientError('Unable to sync with guest agent after {} tries.'.
                          format(n))

    def __enter__(self):
        self.client = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.client.settimeout(self.timeout)
        self.client.connect('/run/qemu.{}.gqa.sock'.format(self.machine))
        self.file = self.client.makefile()
        fcntl.flock(self.file.fileno(), fcntl.LOCK_EX)
        self.sync()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.file.close()
        self.client.close()



from setuptools import setup, find_packages

with open('README.txt') as f:
    readme = f.read() + '\n'
with open('CHANGES.txt') as f:
    readme += f.read() + '\n'
with open('HACKING.txt') as f:
    readme += f.read()

setup(
    name='fc.qemu',
    version='1.0.3',
    author='Christian Kauhaus, Christian Theune',
    author_email='mail@flyingcircus.io',
    url='http://bitbucket.org/flyingcircus/fc.qemu',
    description='Qemu VM management utilities',
    long_description=readme,
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Programming Language :: Python :: 2.7',
    ],
    zip_safe=False,
    license='BSD',
    namespace_packages=['fc'],
    install_requires=[
        'colorama',
        'consulate>=0.4',
        'psutil>=2.1',
        'PyYaml>=3.10',
        'requests',
        'setuptools',
        'structlog',
    ],
    entry_points={
        'console_scripts': [
            'fc-qemu = fc.qemu.main:main',
        ],
    },
)



from ..volume import Volume
import os.path
import pytest
import rbd
import time


@pytest.yield_fixture
def volume(ceph_inst):
    volume = Volume(ceph_inst, 'othervolume', 'label')
    volume.snapshots.purge()
    try:
        if volume._image:
            volume._image.close()
        rbd.RBD().remove(ceph_inst.ioctx, 'othervolume')
    except rbd.ImageNotFound:
        pass

    yield volume
    time.sleep(.2)

    lock = volume.lock_status()
    if lock is not None:
        volume.rbdimage.break_lock(*lock)
    volume.snapshots.purge()
    try:
        if volume._image:
            volume._image.close()
        rbd.RBD().remove(ceph_inst.ioctx, 'othervolume')
    except rbd.ImageNotFound:
        pass


def test_volume_presence(volume):
    assert volume.fullname == 'rbd.hdd/othervolume'
    assert not volume.exists()
    with pytest.raises(rbd.ImageNotFound):
        volume.rbdimage
    volume.ensure_presence()
    assert volume.rbdimage
    # Check that ensure_presence is fine with being called multiple times.
    volume.ensure_presence()


def test_volume_snapshot(volume):
    volume.ensure_presence()
    volume.snapshots.create('s0')
    snaps = list(volume.snapshots)
    assert len(snaps) == 1
    snapshot = snaps[0]
    assert snapshot.name == 'othervolume'
    assert snapshot.snapname == 's0'
    assert snapshot.size == volume.size
    assert snapshot.id == volume.snapshots['s0'].id

    snapshot.remove()
    assert [] == list(volume.snapshots)


def test_purge_snapshots(volume):
    volume.ensure_presence()
    for snap in ['s0', 's1']:
        volume.snapshots.create(snap)
    assert len(volume.snapshots) == 2
    volume.snapshots.purge()
    assert len(volume.snapshots) == 0


def test_snapshot_not_found(volume):
    with pytest.raises(KeyError):
        volume.snapshots['no-such-key']


def test_volume_size(volume):
    volume.ensure_presence()
    assert volume.rbdimage
    assert volume.size == 1024
    volume.ensure_size(2048)
    assert volume.size == 2048
    # Call ensure multiple times to help triggering caching code paths.
    volume.ensure_size(2048)
    assert volume.size == 2048


def test_volume_shared_lock_protection(volume):
    volume.ensure_presence()
    volume.rbdimage.lock_shared('host1', 'a')
    volume.rbdimage.lock_shared('remotehost', 'a')
    with pytest.raises(NotImplementedError):
        volume.lock_status()
    lockers = volume.rbdimage.list_lockers()
    for client, cookie, _ in lockers['lockers']:
        volume.rbdimage.break_lock(client, cookie)


def test_volume_locking(volume):
    # Non-existing volumes report None as locking status but do not raise
    # an exception.
    assert not volume.exists()
    assert volume.lock_status() is None
    volume.ensure_presence()
    assert volume.lock_status() is None
    volume.lock()
    assert volume.lock_status()[1] == 'host1'
    # We want to smoothen out that some other process has locked the same image
    # for the same tag already and assume that this is another incarnation of
    # us - for that we have our own lock.
    volume.lock()
    assert volume.lock_status()[1] == 'host1'
    volume.unlock()
    assert volume.lock_status() is None
    # We can call unlock twice if it isn't locked.
    volume.unlock()

    volume.rbdimage.lock_exclusive('someotherhost')
    with pytest.raises(rbd.ImageBusy):
        volume.lock()

    # Can not unlock locks that someone else holds.
    volume.unlock()
    assert volume.lock_status()[1] == 'someotherhost'


def test_force_unlock(volume):
    volume.ensure_presence()
    volume.rbdimage.lock_exclusive('someotherhost')
    volume.unlock(force=True)
    assert volume.lock_status() is None


def test_volume_mkswap(volume):
    volume.ensure_presence()
    volume.ensure_size(5 * 1024 ** 2)
    with volume.mapped():
        volume.mkswap()


def test_volume_mkfs(volume):
    volume.ensure_presence()
    volume.ensure_size(40 * 1024 ** 2)
    with volume.mapped():
        volume.mkfs()


def test_unmapped_volume_should_have_no_part1(volume):
    # not something like 'None-part1'
    assert volume.part1dev is None


def test_volume_map_unmap_is_idempotent(volume):
    volume.ensure_presence()
    volume.map()
    assert os.path.exists('/dev/rbd/rbd.hdd/othervolume')
    volume.map()
    assert os.path.exists('/dev/rbd/rbd.hdd/othervolume')
    volume.unmap()
    assert not os.path.exists('/dev/rbd/rbd.hdd/othervolume')
    volume.unmap()
    assert not os.path.exists('/dev/rbd/rbd.hdd/othervolume')


def test_map_snapshot(volume):
    volume.ensure_presence()
    volume.snapshots.create('s0')
    with volume.snapshots['s0'].mapped() as device:
        assert os.path.exists(device)


def test_mount_should_fail_if_not_mapped(volume):
    volume.ensure_presence()
    with pytest.raises(RuntimeError):
        volume.mount()


def test_mount_snapshot(volume):
    volume.ensure_presence()
    volume.ensure_size(40 * 1024 ** 2)
    with volume.mapped():
        volume.mkfs(fstype='xfs', gptbios=True)
    volume.snapshots.create('s0')
    snap = volume.snapshots['s0']
    with snap.mounted() as mp:
        mountpoint = mp
        assert os.path.ismount(mp)
        with open('/proc/self/mounts') as mounts:
            assert '{} xfs ro'.format(mp) in mounts.read()
        # test for idempotence
        snap.mount()
        assert os.path.ismount(mp)
    assert not os.path.ismount(mountpoint)
    # test for idempotence
    snap.umount()
    assert not os.path.ismount(mountpoint)



from fc.qemu.logging import prefix, MultiOptimisticLoggerFactory
from structlog import PrintLoggerFactory


def test_prefix():
    assert prefix("test00", "asdfasdfasdfa") == """\
test00>\tasdfasdfasdfa\
"""
    assert prefix("test00", "first line\nsecond line\nthird line\n") == """\
test00>\tfirst line
test00>\tsecond line
test00>\tthird line
test00>\t"""


def test_multi_logger_ignores_error_on_fd(tmpdir):
    closed_stdout = open(str(tmpdir / 'stdout'), 'w')
    closed_stdout.close()
    f = MultiOptimisticLoggerFactory(console=PrintLoggerFactory(closed_stdout))
    logger = f()
    logger.msg(console='asdf')



"""Global helper functions and utilites for fc.qemu."""

from __future__ import print_function
from structlog import get_logger
import contextlib
import filecmp
import os
import subprocess
import sys
import tempfile
import time
import datetime

MiB = 2 ** 20
GiB = 2 ** 30


log = get_logger()


class ControlledRuntimeException(RuntimeError):
    """An exception that is used for flow control but doesn't have to be logged
    as it is handled properly inside.
    """


@contextlib.contextmanager
def rewrite(filename):
    """Rewrite an existing file atomically.

    Clients are allowed to delete the tmpfile to signal that they don't
    want to have it updated.
    """

    with tempfile.NamedTemporaryFile(
            dir=os.path.dirname(filename), delete=False,
            prefix=os.path.basename(filename) + '.') as tf:
        if os.path.exists(filename):
            os.chmod(tf.name, os.stat(filename).st_mode & 0o7777)
        tf.has_changed = False
        yield tf
        if not os.path.exists(tf.name):
            return
        filename_tmp = tf.name
    if (os.path.exists(filename) and
            filecmp.cmp(filename, filename_tmp, shallow=False)):
        os.unlink(filename_tmp)
    else:
        os.rename(filename_tmp, filename)
        tf.has_changed = True


def parse_address(addr):
    if addr.startswith('['):
        host, port = addr[1:].split(']:')
    else:
        host, port = addr.split(':')
    return host, int(port)


def locate_live_service(consul, service_id):
    """Locate Consul service with at least one passing health check.

    It is an error if multiple live services with the same service name
    are found.
    """
    def passing(checks):
        return (any(check['Status'] == 'passing' for check in checks) and
                not any(check['Status'] == 'critical' for check in checks))

    live = [svc for svc in consul.health.service(service_id)
            if passing(svc['Checks'])]
    if len(live) > 1:
        raise RuntimeError('multiple services with passing checks found',
                           service_id)
    return live[0]['Service'] if len(live) else None


def remove_empty_dirs(d):
    """Remove all empty directories from d up.

    Stops on the first non-empty directory.
    """
    while d != '/':
        try:
            os.rmdir(d)
        except OSError:
            return
        d = os.path.dirname(d)


def cmd(cmdline, log):
    """Execute cmdline with stdin closed to avoid questions on terminal"""
    prefix = cmdline.split()[0]
    args = " ".join(cmdline.split()[1:])
    log.debug(prefix, args=args)
    with open('/dev/null') as null:
        output = subprocess.check_output(
            cmdline, shell=True, stdin=null, stderr=subprocess.STDOUT)
    output = output.strip()
    # Keep this here for compatibility with tests
    if output:
        log.debug(prefix, output=output)
    return output


@contextlib.contextmanager
def timeit(label):
    start = time.time()
    yield
    print('run time for {}: {}'.format(label, time.time() - start),
          file=sys.stderr)


def today():
    return datetime.date.today()






from ..agent import swap_size, tmp_size, Agent
from ..conftest import get_log
from ..ellipsis import Ellipsis
from ..util import MiB, GiB
from fc.qemu import util
import datetime
import os
import os.path
import pytest
import shutil
import subprocess


def test_simple_vm_lifecycle_start_stop(vm):
    util.test_log_options['show_events'] = ['vm-status', 'rbd-status']

    vm.status()

    status = get_log()
    assert status == """\
event=vm-status machine=simplevm result=offline
event=rbd-status locker=None machine=simplevm volume=rbd.ssd/simplevm.root
event=rbd-status locker=None machine=simplevm volume=rbd.ssd/simplevm.swap
event=rbd-status locker=None machine=simplevm volume=rbd.ssd/simplevm.tmp"""

    util.test_log_options['show_events'] = []
    vm.start()

    out = get_log()
    # This is 1 end-to-end logging test to see everything.
    assert out == Ellipsis("""\
event=acquire-lock machine=simplevm target=/run/qemu.simplevm.lock
event=acquire-lock machine=simplevm result=locked target=/run/qemu.simplevm.lock
count=1 event=lock-status machine=simplevm
event=generate-config machine=simplevm
event=ensure-root machine=simplevm subsystem=ceph
event=create-vm machine=simplevm subsystem=ceph
args=-I rbd.ssd simplevm event=/usr/local/sbin/create-vm machine=simplevm subsystem=ceph
event=lock machine=simplevm subsystem=ceph volume=rbd.ssd/simplevm.root
event=ensure-tmp machine=simplevm subsystem=ceph
event=lock machine=simplevm subsystem=ceph volume=rbd.ssd/simplevm.tmp
args=-c "/etc/ceph/ceph.conf" --id "admin" map "rbd.ssd/simplevm.tmp" event=rbd machine=simplevm subsystem=ceph volume=rbd.ssd/simplevm.tmp
event=create-fs machine=simplevm subsystem=ceph type=xfs volume=rbd.ssd/simplevm.tmp
args=-o "/dev/rbd/rbd.ssd/simplevm.tmp" event=sgdisk machine=simplevm subsystem=ceph volume=rbd.ssd/simplevm.tmp
event=sgdisk machine=simplevm output=Creating new GPT entries.
The operation has completed successfully. subsystem=ceph volume=rbd.ssd/simplevm.tmp
args=-a 8192 -n 1:8192:0 -c "1:tmp" -t 1:8300 "/dev/rbd/rbd.ssd/simplevm.tmp" event=sgdisk machine=simplevm subsystem=ceph volume=rbd.ssd/simplevm.tmp
event=sgdisk machine=simplevm output=The operation has completed successfully. subsystem=ceph volume=rbd.ssd/simplevm.tmp
args=/dev/rbd/rbd.ssd/simplevm.tmp event=partprobe machine=simplevm subsystem=ceph volume=rbd.ssd/simplevm.tmp
args=-q -f -L "tmp" "/dev/rbd/rbd.ssd/simplevm.tmp-part1" event=mkfs.xfs machine=simplevm subsystem=ceph volume=rbd.ssd/simplevm.tmp
event=mkfs.xfs machine=simplevm output=log stripe unit (4194304 bytes) is too large (maximum is 256KiB)
log stripe unit adjusted to 32KiB subsystem=ceph volume=rbd.ssd/simplevm.tmp
event=seed machine=simplevm subsystem=ceph volume=rbd.ssd/simplevm.tmp
args="/dev/rbd/rbd.ssd/simplevm.tmp-part1" "/mnt/rbd/rbd.ssd/simplevm.tmp" event=mount machine=simplevm subsystem=ceph volume=rbd.ssd/simplevm.tmp
args="/mnt/rbd/rbd.ssd/simplevm.tmp" event=umount machine=simplevm subsystem=ceph volume=rbd.ssd/simplevm.tmp
args=-c "/etc/ceph/ceph.conf" --id "admin" unmap "/dev/rbd/rbd.ssd/simplevm.tmp" event=rbd machine=simplevm subsystem=ceph volume=rbd.ssd/simplevm.tmp
event=ensure-swap machine=simplevm subsystem=ceph
event=lock machine=simplevm subsystem=ceph volume=rbd.ssd/simplevm.swap
args=-c "/etc/ceph/ceph.conf" --id "admin" map "rbd.ssd/simplevm.swap" event=rbd machine=simplevm subsystem=ceph volume=rbd.ssd/simplevm.swap
args=-f -L "swap" "/dev/rbd/rbd.ssd/simplevm.swap" event=mkswap machine=simplevm subsystem=ceph volume=rbd.ssd/simplevm.swap
event=mkswap machine=simplevm output=Setting up swapspace version 1, size = 1048572 KiB
LABEL=swap, UUID=...-...-...-...-... subsystem=ceph volume=rbd.ssd/simplevm.swap
args=-c "/etc/ceph/ceph.conf" --id "admin" unmap "/dev/rbd/rbd.ssd/simplevm.swap" event=rbd machine=simplevm subsystem=ceph volume=rbd.ssd/simplevm.swap
event=acquire-global-lock machine=simplevm subsystem=qemu target=/run/fc-qemu.lock
event=global-lock-acquire machine=simplevm result=locked subsystem=qemu target=/run/fc-qemu.lock
count=1 event=global-lock-status machine=simplevm subsystem=qemu target=/run/fc-qemu.lock
available_real=... bookable=0 event=sufficient-host-memory machine=simplevm required=384 subsystem=qemu
event=start-qemu machine=simplevm subsystem=qemu
additional_args=() event=qemu-system-x86_64 local_args=[\'-nodefaults\', \'-name simplevm,process=kvm.simplevm\', \'-chroot /srv/vm/simplevm\', \'-runas nobody\', \'-serial file:/var/log/vm/simplevm.log\', \'-display vnc=host1:2345\', \'-pidfile /run/qemu.simplevm.pid\', \'-vga std\', \'-m 256\', \'-readconfig /run/qemu.simplevm.cfg\'] machine=simplevm subsystem=qemu
count=0 event=global-lock-status machine=simplevm subsystem=qemu target=/run/fc-qemu.lock
event=global-lock-release machine=simplevm subsystem=qemu target=/run/fc-qemu.lock
event=global-lock-release machine=simplevm result=unlocked subsystem=qemu
arguments={} event=qmp_capabilities id=None machine=simplevm subsystem=qemu/qmp
arguments={} event=query-status id=None machine=simplevm subsystem=qemu/qmp
event=consul-register machine=simplevm
arguments={} event=query-block id=None machine=simplevm subsystem=qemu/qmp
action=throttle current_iops=0 device=virtio0 event=ensure-throttle machine=simplevm target_iops=3000
arguments={\'bps_rd\': 0, \'bps_wr\': 0, \'bps\': 0, \'iops\': 3000, \'iops_rd\': 0, \'device\': u\'virtio0\', \'iops_wr\': 0} event=block_set_io_throttle id=None machine=simplevm subsystem=qemu/qmp
action=throttle current_iops=0 device=virtio1 event=ensure-throttle machine=simplevm target_iops=3000
arguments={\'bps_rd\': 0, \'bps_wr\': 0, \'bps\': 0, \'iops\': 3000, \'iops_rd\': 0, \'device\': u\'virtio1\', \'iops_wr\': 0} event=block_set_io_throttle id=None machine=simplevm subsystem=qemu/qmp
action=throttle current_iops=0 device=virtio2 event=ensure-throttle machine=simplevm target_iops=3000
arguments={\'bps_rd\': 0, \'bps_wr\': 0, \'bps\': 0, \'iops\': 3000, \'iops_rd\': 0, \'device\': u\'virtio2\', \'iops_wr\': 0} event=block_set_io_throttle id=None machine=simplevm subsystem=qemu/qmp
action=none event=ensure-watchdog machine=simplevm
arguments={'command-line': 'watchdog_action action=none'} event=human-monitor-command id=None machine=simplevm subsystem=qemu/qmp
count=0 event=lock-status machine=simplevm
event=release-lock machine=simplevm target=/run/qemu.simplevm.lock
event=release-lock machine=simplevm result=unlocked target=/run/qemu.simplevm.lock""")

    util.test_log_options['show_events'] = [
        'vm-status', 'rbd-status', 'disk-throttle']

    vm.status()
    assert get_log() == Ellipsis("""\
event=vm-status machine=simplevm result=online
device=virtio0 event=disk-throttle iops=3000 machine=simplevm
device=virtio1 event=disk-throttle iops=3000 machine=simplevm
device=virtio2 event=disk-throttle iops=3000 machine=simplevm
event=rbd-status locker=('client...', 'host1') machine=simplevm volume=rbd.ssd/simplevm.root
event=rbd-status locker=('client...', 'host1') machine=simplevm volume=rbd.ssd/simplevm.swap
event=rbd-status locker=('client...', 'host1') machine=simplevm volume=rbd.ssd/simplevm.tmp""")

    vm.stop()
    get_log()

    vm.status()
    assert get_log() == """\
event=vm-status machine=simplevm result=offline
event=rbd-status locker=None machine=simplevm volume=rbd.ssd/simplevm.root
event=rbd-status locker=None machine=simplevm volume=rbd.ssd/simplevm.swap
event=rbd-status locker=None machine=simplevm volume=rbd.ssd/simplevm.tmp"""


def test_simple_vm_lifecycle_ensure_going_offline(vm, capsys, caplog):
    util.test_log_options['show_events'] = ['vm-status', 'rbd-status', 'ensure-state', 'disk-throttle']
    vm.status()
    assert get_log() == """\
event=vm-status machine=simplevm result=offline
event=rbd-status locker=None machine=simplevm volume=rbd.ssd/simplevm.root
event=rbd-status locker=None machine=simplevm volume=rbd.ssd/simplevm.swap
event=rbd-status locker=None machine=simplevm volume=rbd.ssd/simplevm.tmp"""

    vm.ensure()
    out = get_log()
    assert "action=start event=ensure-state found=offline machine=simplevm wanted=online" in out

    vm.status()
    assert get_log() == Ellipsis("""\
event=vm-status machine=simplevm result=online
device=virtio0 event=disk-throttle iops=3000 machine=simplevm
device=virtio1 event=disk-throttle iops=3000 machine=simplevm
device=virtio2 event=disk-throttle iops=3000 machine=simplevm
event=rbd-status locker=('client...', 'host1') machine=simplevm volume=rbd.ssd/simplevm.root
event=rbd-status locker=('client...', 'host1') machine=simplevm volume=rbd.ssd/simplevm.swap
event=rbd-status locker=('client...', 'host1') machine=simplevm volume=rbd.ssd/simplevm.tmp""")

    vm.enc['parameters']['online'] = False
    vm.enc['consul-generation'] += 1
    vm.stage_new_config()
    # As we're re-using the same agent object, we have to time-travel here,
    # otherwise ensure will already think we're on the new generation.
    vm.enc['consul-generation'] -= 1
    vm.ensure()
    assert get_log() == """\
action=stop event=ensure-state found=online machine=simplevm wanted=offline"""

    vm.status()
    assert get_log() == """\
event=vm-status machine=simplevm result=offline
event=rbd-status locker=None machine=simplevm volume=rbd.ssd/simplevm.root
event=rbd-status locker=None machine=simplevm volume=rbd.ssd/simplevm.swap
event=rbd-status locker=None machine=simplevm volume=rbd.ssd/simplevm.tmp"""


def test_vm_not_running_here(vm, capsys):
    util.test_log_options['show_events'] = [
        'vm-status', 'rbd-status']

    vm.status()
    assert get_log() == """\
event=vm-status machine=simplevm result=offline
event=rbd-status locker=None machine=simplevm volume=rbd.ssd/simplevm.root
event=rbd-status locker=None machine=simplevm volume=rbd.ssd/simplevm.swap
event=rbd-status locker=None machine=simplevm volume=rbd.ssd/simplevm.tmp"""

    vm.enc['parameters']['kvm_host'] = 'otherhost'
    vm.enc['consul-generation'] += 1
    vm.stage_new_config()
    vm.enc['consul-generation'] -= 1
    vm.ensure()
    vm.status()
    assert get_log() == Ellipsis("""\
event=vm-status machine=simplevm result=offline
event=rbd-status locker=None machine=simplevm volume=rbd.ssd/simplevm.root
event=rbd-status locker=None machine=simplevm volume=rbd.ssd/simplevm.swap
event=rbd-status locker=None machine=simplevm volume=rbd.ssd/simplevm.tmp""")


def test_crashed_vm_clean_restart(vm):
    util.test_log_options['show_events'] = [
        'rbd-status', 'vm-status', 'ensure', 'throttle', 'shutdown']

    vm.status()

    assert get_log() == Ellipsis("""\
event=vm-status machine=simplevm result=offline
event=rbd-status locker=None machine=simplevm volume=rbd.ssd/simplevm.root
event=rbd-status locker=None machine=simplevm volume=rbd.ssd/simplevm.swap
event=rbd-status locker=None machine=simplevm volume=rbd.ssd/simplevm.tmp""")

    vm.ensure()
    vm.status()
    assert get_log() == Ellipsis("""\
event=running-ensure generation=0 machine=simplevm
action=start event=ensure-state found=offline machine=simplevm wanted=online
...
event=vm-status machine=simplevm result=online
device=virtio0 event=disk-throttle iops=3000 machine=simplevm
device=virtio1 event=disk-throttle iops=3000 machine=simplevm
device=virtio2 event=disk-throttle iops=3000 machine=simplevm
event=rbd-status locker=('client...', 'host1') machine=simplevm volume=rbd.ssd/simplevm.root
event=rbd-status locker=('client...', 'host1') machine=simplevm volume=rbd.ssd/simplevm.swap
event=rbd-status locker=('client...', 'host1') machine=simplevm volume=rbd.ssd/simplevm.tmp""")

    vm.qemu.proc().kill()
    vm.qemu.proc().wait(2)
    get_log()

    vm.status()
    assert get_log() == Ellipsis("""\
event=vm-status machine=simplevm result=offline
event=rbd-status locker=('client...', 'host1') machine=simplevm volume=rbd.ssd/simplevm.root
event=rbd-status locker=('client...', 'host1') machine=simplevm volume=rbd.ssd/simplevm.swap
event=rbd-status locker=('client...', 'host1') machine=simplevm volume=rbd.ssd/simplevm.tmp""")

    vm.ensure()

    vm.status()
    assert get_log() == Ellipsis("""\
event=running-ensure generation=0 machine=simplevm
action=start event=ensure-state found=offline machine=simplevm wanted=online
...
event=vm-status machine=simplevm result=online
device=virtio0 event=disk-throttle iops=3000 machine=simplevm
device=virtio1 event=disk-throttle iops=3000 machine=simplevm
device=virtio2 event=disk-throttle iops=3000 machine=simplevm
event=rbd-status locker=('client...', 'host1') machine=simplevm volume=rbd.ssd/simplevm.root
event=rbd-status locker=('client...', 'host1') machine=simplevm volume=rbd.ssd/simplevm.swap
event=rbd-status locker=('client...', 'host1') machine=simplevm volume=rbd.ssd/simplevm.tmp""")

    util.test_log_options['show_events'] = [
        'shutdown', 'kill', 'unlock', 'vm-status', 'consul', 'clean', 'rbd-status']
    vm.stop()
    vm.status()
    assert get_log() == Ellipsis("""\
event=graceful-shutdown machine=simplevm
event=graceful-shutdown-failed machine=simplevm reason=timeout
event=kill-vm machine=simplevm
event=killed-vm machine=simplevm
event=unlock machine=simplevm subsystem=ceph volume=rbd.ssd/simplevm.root
event=unlock machine=simplevm subsystem=ceph volume=rbd.ssd/simplevm.swap
event=unlock machine=simplevm subsystem=ceph volume=rbd.ssd/simplevm.tmp
event=consul-deregister machine=simplevm
event=clean-run-files machine=simplevm subsystem=qemu
event=vm-status machine=simplevm result=offline
event=rbd-status locker=None machine=simplevm volume=rbd.ssd/simplevm.root
event=rbd-status locker=None machine=simplevm volume=rbd.ssd/simplevm.swap
event=rbd-status locker=None machine=simplevm volume=rbd.ssd/simplevm.tmp
address=host1 event=consul machine=simplevm service=qemu-simplevm""")


def test_do_not_clean_up_crashed_vm_that_doesnt_get_restarted(vm):
    vm.ensure()
    assert vm.qemu.is_running() is True
    vm.qemu.proc().kill()
    vm.qemu.proc().wait(2)
    assert vm.ceph.locked_by_me() is True
    vm.enc['parameters']['online'] = False
    vm.enc['consul-generation'] += 1
    vm.stage_new_config()
    vm.enc['consul-generation'] -= 1
    vm.ensure()
    # We don't really know what's going on here, so, yeah, don't touch it.
    assert vm.ceph.locked_by_me() is True


def test_vm_snapshot_only_if_running(vm):
    assert list(x.fullname for x in vm.ceph.root.snapshots) == []
    vm.ceph.root.ensure_presence()
    vm.snapshot('asdf')
    assert list(x.fullname for x in vm.ceph.root.snapshots) == []


def test_vm_snapshot_with_missing_guest_agent(vm, monkeypatch):
    util.test_log_options['show_events'] = [
        'consul', 'snapshot', 'freeze', 'thaw']

    monkeypatch.setattr(
        util, 'today', lambda: datetime.date(2010, 1, 1))

    assert list(x.fullname for x in vm.ceph.root.snapshots) == []
    vm.ensure()
    get_log()

    with pytest.raises(Exception):
        vm.snapshot('asdf', 7)
    assert Ellipsis("""\
event=snapshot-create machine=simplevm name=asdf-keep-until-20100108
event=freeze machine=simplevm volume=root
action=continue event=freeze-failed machine=simplevm reason=Unable to sync \
with guest agent after 20 tries.
event=snapshot-ignore machine=simplevm reason=not frozen
event=thaw machine=simplevm volume=root
action=retry event=thaw-failed machine=simplevm reason=Unable to sync with \
guest agent after 20 tries.
action=continue event=thaw-failed machine=simplevm reason=Unable to sync \
with guest agent after 20 tries.""") == get_log()

    with pytest.raises(Exception):
        vm.snapshot('asdf', 0)
    assert """\
event=snapshot-create machine=simplevm name=asdf
event=freeze machine=simplevm volume=root
action=continue event=freeze-failed machine=simplevm reason=[Errno 11] Resource temporarily unavailable
event=snapshot-ignore machine=simplevm reason=not frozen
event=thaw machine=simplevm volume=root
action=retry event=thaw-failed machine=simplevm reason=[Errno 11] Resource temporarily unavailable
action=continue event=thaw-failed machine=simplevm reason=[Errno 11] Resource temporarily unavailable\
""" == get_log()


def test_vm_throttle_iops(vm):
    vm.start()
    get_log()

    vm.ensure_online_disk_throttle()
    assert get_log() == """\
arguments={} event=query-block id=None machine=simplevm subsystem=qemu/qmp
action=none current_iops=3000 device=virtio0 event=ensure-throttle machine=simplevm target_iops=3000
action=none current_iops=3000 device=virtio1 event=ensure-throttle machine=simplevm target_iops=3000
action=none current_iops=3000 device=virtio2 event=ensure-throttle machine=simplevm target_iops=3000"""

    vm.cfg['iops'] = 10

    vm.ensure_online_disk_throttle()
    assert get_log() == """\
arguments={} event=query-block id=None machine=simplevm subsystem=qemu/qmp
action=throttle current_iops=3000 device=virtio0 event=ensure-throttle machine=simplevm target_iops=10
arguments={'bps_rd': 0, 'bps_wr': 0, 'bps': 0, 'iops': 10, 'iops_rd': 0, 'device': u'virtio0', 'iops_wr': 0} event=block_set_io_throttle id=None machine=simplevm subsystem=qemu/qmp
action=throttle current_iops=3000 device=virtio1 event=ensure-throttle machine=simplevm target_iops=10
arguments={'bps_rd': 0, 'bps_wr': 0, 'bps': 0, 'iops': 10, 'iops_rd': 0, 'device': u'virtio1', 'iops_wr': 0} event=block_set_io_throttle id=None machine=simplevm subsystem=qemu/qmp
action=throttle current_iops=3000 device=virtio2 event=ensure-throttle machine=simplevm target_iops=10
arguments={'bps_rd': 0, 'bps_wr': 0, 'bps': 0, 'iops': 10, 'iops_rd': 0, 'device': u'virtio2', 'iops_wr': 0} event=block_set_io_throttle id=None machine=simplevm subsystem=qemu/qmp"""

    vm.ensure_online_disk_throttle()
    assert get_log() == """\
arguments={} event=query-block id=None machine=simplevm subsystem=qemu/qmp
action=none current_iops=10 device=virtio0 event=ensure-throttle machine=simplevm target_iops=10
action=none current_iops=10 device=virtio1 event=ensure-throttle machine=simplevm target_iops=10
action=none current_iops=10 device=virtio2 event=ensure-throttle machine=simplevm target_iops=10"""


def test_vm_resize_disk(vm):
    vm.start()
    get_log()
    vm.ensure_online_disk_size()
    assert get_log() == """\
action=none event=check-disk-size found=5368709120 machine=simplevm wanted=5368709120\
"""

    vm.cfg['disk'] *= 2

    vm.ensure_online_disk_size()
    assert get_log() == """\
action=resize event=check-disk-size found=5368709120 machine=simplevm wanted=10737418240
arguments={'device': 'virtio0', 'size': 10737418240} event=block_resize id=None machine=simplevm subsystem=qemu/qmp"""

    vm.ensure_online_disk_size()
    assert get_log() == """\
action=none event=check-disk-size found=10737418240 machine=simplevm wanted=10737418240"""


def test_swap_size():
    assert swap_size(512) == 1024 * MiB
    assert swap_size(768) == 1024 * MiB
    assert swap_size(1024) == 1024 * MiB
    assert swap_size(2048) == 1448 * MiB
    assert swap_size(4096) == 2048 * MiB
    assert swap_size(8192) == 2896 * MiB


def test_tmp_size():
    assert tmp_size(30) == 5 * GiB
    assert tmp_size(50) == 7 * GiB
    assert tmp_size(100) == 10 * GiB
    assert tmp_size(500) == 22 * GiB
    assert tmp_size(1000) == 31 * GiB


def test_vm_migration(vm):
    subprocess.check_call('./test-migration.sh', shell=True)


def test_simple_cancelled_migration_doesnt_clean_up(vm, monkeypatch):
    import fc.qemu.outgoing
    monkeypatch.setattr(fc.qemu.outgoing.Outgoing, 'connect_timeout', 2)

    vm.start()
    assert os.path.exists('/run/qemu.simplevm.pid')
    assert vm.ceph.locked_by_me()

    vm.ensure_online_remote()
    assert vm.ceph.locked_by_me()
    assert os.path.exists('/run/qemu.simplevm.pid')


def test_new_vm(vm):
    # A new VM gets created by consul adding the staging filename and then
    # starting it. At this point the main config file doesn't exist yet.
    shutil.copy('/etc/qemu/vm/simplevm.cfg',
                '/etc/qemu/vm/.simplevm.cfg.staging')
    os.unlink('/etc/qemu/vm/simplevm.cfg')
    # Include testing the agent setup for this scenario.
    vm = Agent('simplevm')
    with vm:
        vm.ensure()
    assert get_log() == Ellipsis("""\
...
event=running-ensure generation=-1 machine=simplevm
...
action=update current=-1 event=update-check machine=simplevm result=update-available update=0
...
action=start event=ensure-state found=offline machine=simplevm wanted=online
...
event=generate-config machine=simplevm
event=ensure-root machine=simplevm subsystem=ceph
event=create-vm machine=simplevm subsystem=ceph
...
""")
